# CAS (SAS Viya) interface for sasoptpy

import inspect

import pandas as pd
import numpy as np
import warnings

import sasoptpy
from sasoptpy.interface import Mediator

class CASMediator(Mediator):
    """
    Handles the connection between sasoptpy and the SAS Viya (CAS) server

    Parameters
    ----------
    caller : :class:`sasoptpy.Model` or :class:`sasoptpy.Workspace`
        Model or workspace that mediator belongs to
    cas_session : :class:`swat.cas.connection.CAS`
        CAS connection

    Notes
    -----

    * CAS Mediator is used by :class:`sasoptpy.Model` and :class:`sasoptpy.Workspace` objects
      internally.

    """

    def __init__(self, caller, cas_session):
        self.caller = caller
        self.session = cas_session

    def solve(self, **kwargs):
        """
        Solve action for :class:`Model` objects
        """
        self.session.loadactionset(actionset='optimization')

        has_user_called_mps = kwargs.get('mps', kwargs.get('frame', False))
        options = kwargs.get('options', dict())

        mps_indicator = self.is_mps_format_needed(
            has_user_called_mps, options)

        if mps_indicator:
            return self.solve_with_mps(**kwargs)
        else:
            return self.solve_with_optmodel(**kwargs)

    def submit(self, **kwargs):
        """
        Submit action for custom input and :class:`sasoptpy.Workspace` objects
        """
        self.session.loadactionset(actionset='optimization')
        return self.submit_optmodel_code(**kwargs)

    def tune(self, **kwargs):
        """
        Wrapper for the MILP tuner
        """
        self.session.loadactionset(actionset='optimization')

        if not hasattr(self.session, 'optimization.tuner'):
            raise RuntimeError('Current CAS session version do not have tuner capability.')

        return self.tune_problem(**kwargs)


    def is_mps_format_needed(self, mps_option, options):

        enforced = False
        mps_option = mps_option
        session = self.session
        caller = self.caller

        model = caller

        # If runOptmodel action is not available on server
        if not hasattr(session.optimization, 'runoptmodel'):
            mps_option = True
            enforced = True

        if 'decomp' in options:
            mps_option = True
            enforced = True

        if mps_option:
            switch = False
            # Sets and parameter belong to abstract models
            if model.get_sets() or model.get_parameters():
                warnings.warn(
                    'INFO: Model {} has abstract elements, '.format(
                        model.get_name()) +
                    'switching to OPTMODEL mode.', UserWarning)
                switch = True
            # MPS format cannot represent nonlinear problems
            elif not sasoptpy.is_linear(model):
                warnings.warn(
                    'INFO: Model {} includes nonlinear or abstract '.format(
                        model.get_name()) +
                    'components, switching to OPTMODEL mode.', UserWarning)
                switch = True

            if switch and enforced and mps_option:
                raise RuntimeError('Problem requires runOptmodel action which '
                                   'is not available or appropriate')
            elif switch:
                mps_option = False

        return mps_option

    def solve_with_mps(self, **kwargs):
        """
        Submits the problem in MPS (DataFrame) format, supported by old versions

        Parameters
        ----------
        kwargs : dict
            Keyword arguments for solver settings and options

        Returns
        -------
        primal_solution : :class:`swat.dataframe.SASDataFrame`
            Solution of the model or None

        """

        model = self.caller
        session = self.session
        verbose = kwargs.get('verbose', False)
        submit = kwargs.get('submit', True)
        options = kwargs.get('options', dict())
        primalin = kwargs.get('primalin', False)
        name = kwargs.get('name', None)
        replace = kwargs.get('replace', True)
        drop = kwargs.get('drop', False)
        user_blocks = None

        print('NOTE: Converting model {} to DataFrame.'.format(model.get_name()))
        # Pre-upload argument parse

        # Find problem type and initial values
        ptype = 1  # LP
        for v in model.get_grouped_variables().values():
            if v._type != sasoptpy.CONT:
                ptype = 2
                break

        # Decomp check
        try:
            if options['decomp']['method'] == 'user':
                        user_blocks = self.upload_user_blocks()
                        options['decomp'] = {'blocks': user_blocks}
        except KeyError:
            pass

        # Initial value check for MIP
        if primalin:
            init_values = []
            var_names = []
            if ptype == 2:
                for v in model.loop_variables():
                    if v._init is not None:
                        var_names.append(v.get_name())
                        init_values.append(v._init)
                if (len(init_values) > 0 and
                   options.get('primalin', 1) is not None):
                    primalinTable = pd.DataFrame(
                        data={'_VAR_': var_names, '_VALUE_': init_values})
                    session.upload_frame(
                        primalinTable, casout={
                            'name': 'PRIMALINTABLE', 'replace': True})
                    options['primalin'] = 'PRIMALINTABLE'

        # Check if objective constant workaround is needed
        sfunc = session.solveLp if ptype == 1 else session.solveMilp
        has_arg = 'objconstant' in inspect.signature(sfunc).parameters
        if has_arg and 'objconstant' not in options:
            objconstant = model.get_objective()._linCoef['CONST']['val']
            options['objconstant'] = objconstant

        # Upload the problem
        mps_table = self.upload_model(name, replace=replace,
                                      constant=not has_arg, verbose=verbose)

        if verbose:
            print(mps_table.to_string())
        if not submit:
            return mps_table

        if ptype == 1:
            valid_opts = inspect.signature(session.solveLp).parameters
            lp_opts = {}
            for key, value in options.items():
                if key in valid_opts:
                    lp_opts[key] = value
            response = session.solveLp(
                data=mps_table.name, **lp_opts,
                primalOut={'caslib': 'CASUSER', 'name': 'primal',
                           'replace': True},
                dualOut={'caslib': 'CASUSER', 'name': 'dual',
                         'replace': True},
                objSense=model.get_objective().get_sense())
        elif ptype == 2:
            valid_opts = inspect.signature(session.solveMilp).parameters
            milp_opts = {}
            for key, value in options.items():
                if key in valid_opts:
                    milp_opts[key] = value
            response = session.solveMilp(
                data=mps_table.name, **milp_opts,
                primalOut={'caslib': 'CASUSER', 'name': 'primal',
                           'replace': True},
                dualOut={'caslib': 'CASUSER', 'name': 'dual',
                         'replace': True},
                objSense=model.get_objective().get_sense())

        model.response = response

        # Parse solution
        if(response.get_tables('status')[0] == 'OK'):
            model._primalSolution = session.CASTable(
                'primal', caslib='CASUSER').to_frame()
            model._dualSolution = session.CASTable(
                'dual', caslib='CASUSER').to_frame()
            # Bring solution to variables
            for _, row in model._primalSolution.iterrows():
                if ('_SOL_' in model._primalSolution and row['_SOL_'] == 1)\
                     or '_SOL_' not in model._primalSolution:
                    model.get_variable(row['_VAR_']).set_value(row['_VALUE_'])

            # Capturing dual values for LP problems
            if ptype == 1:
                model._primalSolution = model._primalSolution[
                    ['_VAR_', '_LBOUND_', '_UBOUND_', '_VALUE_',
                     '_R_COST_']]
                model._primalSolution.columns = ['var', 'lb', 'ub',
                                                'value', 'rc']
                model._dualSolution = model._dualSolution[
                    ['_ROW_', '_ACTIVITY_', '_VALUE_']]
                model._dualSolution.columns = ['con', 'value', 'dual']
                for row in model._primalSolution.itertuples():
                    model.get_variable(row.var)._dual = row.rc
                for row in model._dualSolution.itertuples():
                    model.get_constraint(row.con)._dual = row.dual
            elif ptype == 2:
                model._primalSolution = model._primalSolution[
                    ['_VAR_', '_LBOUND_', '_UBOUND_', '_VALUE_',
                     '_SOL_']]
                model._primalSolution.columns = ['var', 'lb', 'ub',
                                                'value', 'solution']
                model._dualSolution = model._dualSolution[
                    ['_ROW_', '_ACTIVITY_', '_SOL_']]
                model._dualSolution.columns = ['con', 'value',
                                              'solution']

        # Drop tables
        if drop:
            session.table.droptable(table=mps_table.name)
            if user_blocks is not None:
                session.table.droptable(table=user_blocks)
            if primalin:
                session.table.droptable(table='PRIMALINTABLE')

        # Post-solve parse
        if(response.get_tables('status')[0] == 'OK'):
            # Print problem and solution summaries
            model._problemSummary = response.ProblemSummary[['Label1',
                                                            'cValue1']]
            model._solutionSummary = response.SolutionSummary[['Label1',
                                                              'cValue1']]
            model._problemSummary.set_index(['Label1'], inplace=True)
            model._problemSummary.columns = ['Value']
            model._problemSummary.index.names = ['Label']
            model._solutionSummary.set_index(['Label1'], inplace=True)
            model._solutionSummary.columns = ['Value']
            model._solutionSummary.index.names = ['Label']
            # Record status and time
            model._status = response.solutionStatus
            model._soltime = response.solutionTime
            if('OPTIMAL' in response.solutionStatus):
                model._objval = response.objective
                # Replace initial values with current values
                for v in model.loop_variables():
                    v._init = v._value
                return model._primalSolution
            else:
                warnings.warn('Solution message is not OPTIMAL: {}'.format(response.solutionStatus), UserWarning)
                model._objval = 0
                return None
        else:
            raise RuntimeError('Solve came back with message: {}'.format(
                response.get_tables('status')[0]))

    def solve_with_optmodel(self, **kwargs):
        """
        Submits the problem in OPTMODEL format

        Parameters
        ----------
        kwargs : dict
            Keyword arguments for solver settings and options

        Returns
        -------
        primal_solution : :class:`swat.dataframe.SASDataFrame`
            Solution of the model or None

        """

        model = self.caller
        session = self.session
        verbose = kwargs.get('verbose', False)
        submit = kwargs.get('submit', True)

        print('NOTE: Converting model {} to OPTMODEL.'.format(
            model.get_name()))
        options = kwargs.get('options', dict())
        primalin = kwargs.get('primalin', False)
        optmodel_string = model.to_optmodel(header=False, options=options,
                                           ods=False, primalin=primalin,
                                            parse=True)
        if verbose:
            print(optmodel_string)
        if not submit:
            return optmodel_string
        print('NOTE: Submitting OPTMODEL code to CAS server.')
        response = session.runOptmodel(
            optmodel_string,
            outputTables={
                'names': {'solutionSummary': 'solutionSummary',
                          'problemSummary': 'problemSummary'}
                }
            )

        model.response = response

        # Parse solution
        return self.parse_cas_solution()

    def parse_cas_solution(self):
        """
        Performs post-solve operations

        Returns
        -------
        solution : :class:`swat.dataframe.SASDataFrame`
            Solution of the problem
        """
        caller = self.caller
        session = self.session
        response = caller.response

        if response.status == 'Syntax Error':
            raise SyntaxError('An invalid symbol is generated, check object names')
        elif response.status == 'Semantic Error':
            raise RuntimeError('A semantic error has occured, check statements and object types')

        solution = session.CASTable('solution').to_frame()
        dual_solution = session.CASTable('dual').to_frame()

        caller._primalSolution = solution
        caller._dualSolution = dual_solution

        caller._problemSummary = self.parse_cas_table('problemSummary')
        caller._solutionSummary = self.parse_cas_table('solutionSummary')

        caller._status = response.solutionStatus
        caller._soltime = response.solutionTime

        self.set_variable_values(solution)
        self.set_constraint_values(dual_solution)

        self.set_model_objective_value()
        self.set_variable_init_values()

        return solution

    def parse_table(self, table):
        session = self.session
        table = session.CASTable(table).to_frame()
        return table

    def parse_cas_table(self, table):
        """
        Converts requested :class:`swat.cas.table.CASTable` objects to
        :class:`swat.dataframe.SASDataFrame`
        """
        session = self.session
        table = session.CASTable(table).to_frame()
        return sasoptpy.interface.parse_optmodel_table(table)

    def set_variable_values(self, solution):
        """
        Performs post-solve assignment of variable values

        Parameters
        ----------
        solution : class:`swat.dataframe.SASDataFrame`
            Primal solution of the problem
        """
        caller = self.caller
        solver = ''
        try:
            solver = caller.get_solution_summary().loc['Solver', 'Value']
        except:
            pass
        for row in solution.itertuples():
            caller.set_variable_value(row.var, row.value)
            if solver == 'LP':
                caller.set_dual_value(row.var, row.rc)

    def set_constraint_values(self, solution):
        """
        Performs post-solve assignment of constraint values

        Parameters
        ----------
        solution : class:`swat.dataframe.SASDataFrame`
            Primal solution of the problem
        """
        caller = self.caller
        solver = ''
        try:
            solver = caller.get_solution_summary().loc['Solver', 'Value']
        except:
            pass
        if solver == 'LP':
            for row in solution.itertuples():
                con = caller.get_constraint(row.con)
                if con is not None:
                    con.set_dual(row.dual)

    def set_model_objective_value(self):
        """
        Performs post-solve assignment of objective values

        Parameters
        ----------
        solution : class:`swat.dataframe.SASDataFrame`
            Primal solution of the problem
        """
        caller = self.caller
        if sasoptpy.core.util.is_model(caller):
            if hasattr(caller.response, 'objective'):
                objval = caller.response.objective
                caller.set_objective_value(objval)

    def set_variable_init_values(self):
        """
        Performs post-solve assignment of variable initial values

        Parameters
        ----------
        solution : class:`swat.dataframe.SASDataFrame`
            Primal solution of the problem
        """
        caller = self.caller
        if sasoptpy.core.util.is_model(caller):
            for v in caller.loop_variables():
                v.set_init(v.get_value())

    def upload_user_blocks(self):
        """
        Uploads user-defined decomposition blocks to the CAS server

        Returns
        -------
        name : string
            CAS table name of the user-defined decomposition blocks

        Examples
        --------

        >>> userblocks = m.upload_user_blocks()
        >>> m.solve(milp={'decomp': {'blocks': userblocks}})

        """
        sess = self.session
        model = self.caller

        blocks_dict = {}
        block_counter = 0
        decomp_table = []
        for c in model.loop_constraints():
            if c._block is not None:
                if c._block not in blocks_dict:
                    blocks_dict[c._block] = block_counter
                    block_counter += 1
                block_no = blocks_dict[c._block]
                decomp_table.append([c.get_name(), block_no])
        frame_decomp_table = pd.DataFrame(decomp_table,
                                          columns=['_ROW_', '_BLOCK_'])
        response = sess.upload_frame(frame_decomp_table,
                                     casout={'name': 'BLOCKSTABLE',
                                             'replace': True})
        return(response.name)

    def upload_model(self, name=None, replace=True, constant=False,
                     verbose=False):
        """
        Converts internal model to MPS table and upload to CAS session

        Parameters
        ----------
        name : string, optional
            Desired name of the MPS table on the server
        replace : boolean, optional
            Option to replace the existing MPS table

        Returns
        -------
        frame : :class:`swat.cas.table.CASTable`
            Reference to the uploaded CAS Table

        Notes
        -----

        - This method returns None if the model session is not valid.
        - Name of the table is randomly assigned if name argument is None
          or not given.
        - This method should not be used if :func:`Model.solve` is going
          to be used. :func:`Model.solve` calls this method internally.

        """
        model = self.caller
        df = model.to_mps(constant=constant)
        if verbose:
            print(df.to_string())

        print('NOTE: Uploading the problem DataFrame to the server.')
        if name is not None:
            return self.session.upload_frame(
                data=df, casout={'name': name, 'replace': replace})
        else:
            return self.session.upload_frame(
                data=df, casout={'replace': replace})

    def submit_optmodel_code(self, **kwargs):
        """
        Converts caller into OPTMODEL code and submits using
        optimization.runOptmodel action

        Parameters
        ----------
        kwargs :
            Solver settings and options
        """

        caller = self.caller
        session = self.session
        optmodel_code = sasoptpy.util.to_optmodel(caller, header=False, parse=True)
        verbose = kwargs.get('verbose', None)

        if verbose:
            print(sasoptpy.to_optmodel(caller))

        response = session.runOptmodel(
            optmodel_code
        )

        caller.response = response
        caller.parse_solve_responses()
        caller.parse_print_responses()
        caller.parse_create_data_responses(self)

        return self.parse_cas_workspace_response()

    def parse_cas_workspace_response(self):
        """
        Parses results of workspace submission
        """
        caller = self.caller
        session = self.session
        response = caller.response

        if response.status == 'Syntax Error':
            raise SyntaxError('An invalid symbol is generated, check object names')
        elif response.status == 'Semantic Error':
            raise RuntimeError('A semantic error has occured, check statements and object types')

        solution = session.CASTable('solution').to_frame()
        dual_solution = session.CASTable('dual').to_frame()

        caller._primalSolution = solution
        caller._dualSolution = dual_solution

        self.set_workspace_variable_values(solution)
        #self.set_constraint_values(dual_solution)

        # self.set_model_objective_value()
        # self.set_variable_init_values()

        return solution

    def set_workspace_variable_values(self, solution):
        """
        Performs post-solve assignment of :class:`sasoptpy.Workspace` variable values
        """
        caller = self.caller
        for row in solution.itertuples():
            caller.set_variable_value(row.var, row.value)

    def tune_problem(self, **kwargs):
        """
        Calls optimization.tuner CAS action to finds out the ideal configuration
        """
        model = self.caller
        session = self.session
        name = model.get_name()

        if not sasoptpy.is_linear(model):
            raise TypeError('Model {} is not linear'.format(model.get_name()))
        if not sasoptpy.util.has_integer_variables(model):
            raise TypeError('Model {} do not have integer or binary variables'.format(model.get_name()))

        self.upload_model(name=name)

        if kwargs.get('tunerParameters') is None:
            kwargs['tunerParameters'] = {'maxconfigs': 100}

        response = session.optimization.tuner(
            instances=[{'data': name}],
            **kwargs
        )

        def replace_column_names(sasdf):
            colnames = sasdf.columns
            collabels = []
            for i in colnames:
                if sasdf.colinfo[i].label is not None:
                    collabels.append(sasdf.colinfo[i].label)
                else:
                    collabels.append(i)
            sasdf.columns = collabels
            return sasdf

        performance = response.PerformanceInformation
        info = response.TunerInformation
        summary = response.TunerSummary
        results = response.TunerResults

        performance = replace_column_names(performance)
        info = replace_column_names(info)
        summary = replace_column_names(summary)
        results = replace_column_names(results)

        model._tunerResults = {
            'Performance Information': performance,
            'Tuner Information': info,
            'Tuner Summary': summary,
            'Tuner Results': results
        }

        return results




import os
from swat import CAS
import warnings

def create_cas_connection():
    cas_args = {'host': os.environ.get('CASHOST'),
                'port': int(os.environ.get('CASPORT'))}
    authinfo = os.environ.get('AUTHINFO', None)
    if authinfo:
        cas_args['authinfo'] = authinfo
        return CAS(**cas_args)

    username = os.environ.get('CASUSERNAME', None)
    if username:
        cas_args['username'] = username
        cas_args['password'] = os.environ.get('CASPASSWORD', None)
        return CAS(**cas_args)

    warnings.warn('CAS connection cannot be established.', RuntimeWarning)
    return None



from .statement_base import Statement

import sasoptpy
from sasoptpy.util.package_utils import _to_sas_string


class LiteralStatement(Statement):

    @sasoptpy.class_containable
    def __init__(self, literal=None, **kwargs):
        super().__init__()
        self.append(literal)

    def append(self, literal):
        self.elements.append(literal)

    def _expr(self):
        expr = '\n'.join(self.elements)
        return expr

    def _defn(self):
        defn = '\n'.join(self.elements)
        return defn

    @classmethod
    def expand(cls):
        es = LiteralStatement('expand;', internal=True)
        return es

    @classmethod
    def union(cls, *args):
        ss = [_to_sas_string(i) for i in args]
        ls = LiteralStatement(' union '.join(ss), internal=True)
        ls.set_internal(True)
        return ls

    @classmethod
    def diff(cls, *args):
        ss = [_to_sas_string(i) for i in args]
        ls = LiteralStatement(' diff '.join(ss), internal=True)
        ls.set_internal(True)
        return ls

    @classmethod
    def substring(cls, main_string, first_pos, last_pos):
        first_el = _to_sas_string(main_string)
        pos = f'{first_pos}, {last_pos}'
        ss = LiteralStatement(f'substr({first_el}, {pos})', internal=True)
        ss.set_internal(True)
        return ss
        #
        # ss = sasoptpy.Auxiliary(base='{}, {}, {}'.format(
        #     _to_sas_string(main_string), first_pos, last_pos
        # ), operator='substr')
        #return ss

    @classmethod
    def use_problem(cls, problem):
        ps = LiteralStatement('use problem {};'.format(
            problem.get_name()
        ))



# Abstract class for sessions

from abc import ABC, abstractmethod

class Mediator(ABC):

    @abstractmethod
    def solve(self, **kwargs):
        pass

    @abstractmethod
    def submit(self, **kwargs):
        pass



import sasoptpy as so
import pandas as pd


def test(cas_conn):

    m = so.Model(name='mining_optimization', session=cas_conn)

    mine_data = pd.DataFrame([
        ['mine1', 5, 2, 1.0],
        ['mine2', 4, 2.5, 0.7],
        ['mine3', 4, 1.3, 1.5],
        ['mine4', 5, 3, 0.5],
        ], columns=['mine', 'cost', 'extract_ub', 'quality']).\
        set_index(['mine'])

    year_data = pd.DataFrame([
        [1, 0.9],
        [2, 0.8],
        [3, 1.2],
        [4, 0.6],
        [5, 1.0],
        ], columns=['year', 'quality_required']).set_index(['year'])

    max_num_worked_per_year = 3
    revenue_per_ton = 10
    discount_rate = 0.10

    MINES = mine_data.index.tolist()
    cost = mine_data['cost']
    extract_ub = mine_data['extract_ub']
    quality = mine_data['quality']
    YEARS = year_data.index.tolist()
    quality_required = year_data['quality_required']

    isOpen = m.add_variables(MINES, YEARS, vartype=so.BIN, name='isOpen')
    isWorked = m.add_variables(MINES, YEARS, vartype=so.BIN, name='isWorked')
    extract = m.add_variables(MINES, YEARS, lb=0, name='extract')
    [extract[i, j].set_bounds(ub=extract_ub[i]) for i in MINES for j in YEARS]

    extractedPerYear = {j: extract.sum('*', j) for j in YEARS}
    discount = {j: 1 / (1+discount_rate) ** (j-1) for j in YEARS}

    totalRevenue = revenue_per_ton *\
        so.expr_sum(discount[j] * extractedPerYear[j] for j in YEARS)
    totalCost = so.expr_sum(discount[j] * cost[i] * isOpen[i, j]
                             for i in MINES for j in YEARS)
    m.set_objective(totalRevenue-totalCost, sense=so.MAX, name='totalProfit')

    m.add_constraints((extract[i, j] <= extract[i, j]._ub * isWorked[i, j]
                      for i in MINES for j in YEARS), name='link')

    m.add_constraints((isWorked.sum('*', j) <= max_num_worked_per_year
                      for j in YEARS), name='cardinality')

    m.add_constraints((isWorked[i, j] <= isOpen[i, j] for i in MINES
                      for j in YEARS), name='worked_implies_open')

    m.add_constraints((isOpen[i, j] <= isOpen[i, j-1] for i in MINES
                      for j in YEARS if j != 1), name='continuity')

    m.add_constraints((so.expr_sum(quality[i] * extract[i, j] for i in MINES)
                      == quality_required[j] * extractedPerYear[j]
                      for j in YEARS), name='quality_con')

    res = m.solve()
    if res is not None:
        print(so.get_solution_table(isOpen, isWorked, extract))
        quality_sol = {j: so.expr_sum(quality[i] * extract[i, j].get_value()
                                       for i in MINES)
                       / extractedPerYear[j].get_value() for j in YEARS}
        qs = so.dict_to_frame(quality_sol, ['quality_sol'])
        epy = so.dict_to_frame(extractedPerYear, ['extracted_per_year'])
        print(so.get_solution_table(epy, qs, quality_required))

    return m.get_objective_value()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Condition for symbolic expressions
"""

import sasoptpy

class Condition:

    def __init__(self, left, c_type, right):
        self._left = left
        self._type = c_type
        self._right = right

    def _expr(self):
        left = self._left._cond_expr() if hasattr(self._left, '_cond_expr') \
            else sasoptpy.to_expression(self._left)

        right = self._right._cond_expr() if hasattr(self._right, '_cond_expr') \
            else sasoptpy.to_expression(self._right)

        if self._type in ['and', 'or', 'AND', 'OR']:
            left = '({})'.format(left)
            right = '({})'.format(right)

        return '{} {} {}'.format(left, self._type, right)

    def _cond_expr(self):
        return self._expr()

    def set_left(self, left):
        self._left = left

    def __and__(self, other):
        r = Condition(left=self, c_type='and', right=other)
        return r

    def __or__(self, other):
        r = Condition(left=self, c_type='or', right=other)
        return r

    def copy(self):
        return Condition(self._left, self._type, self._right)


class Conditional:

    def __init__(self, parent, format=None):
        self._parent = parent
        self._conditions = []

    def under_condition(self, condition):
        self.add_condition(condition)
        return True

    def add_condition(self, condition):
        if sasoptpy.core.util.is_constraint(condition):
            self.add_exp_condition(condition)
        elif isinstance(condition, Condition):
            self._conditions.append(condition)

    def add_exp_condition(self, condition):
        self._conditions.append(condition)

    def add_custom_condition(self, operation, key):
        c = Condition(left=self._parent, c_type=operation, right=key)
        self._conditions.append(c)
        return c

    def get_conditions_len(self):
        return len(self._conditions)

    def get_conditions(self):
        return self._conditions

    def get_conditions_str(self):
        keyword_list = [' and ', ' or ']
        conds = [sasoptpy.util.to_condition_expression(c) for c in self._conditions]
        if len(conds) > 0:
            for i, c in enumerate(conds):
                if any(keyword in c for keyword in keyword_list):
                    conds[i] = '(' + c + ')'
            return ' and '.join(conds)
        else:
            return ''

    def copy_conditions(self, other):
        for i in other.get_conditions():
            con_copy = i.copy()
            con_copy.set_left(self._parent)
            self._conditions.append(con_copy)

    def is_in(self, set):
        c = self.add_custom_condition('IN', set)
        return c

    def __eq__(self, key):
        c = self.add_custom_condition('EQ', key)
        return c

    def __le__(self, key):
        c = self.add_custom_condition('<=', key)
        return c

    def __lt__(self, key):
        c = self.add_custom_condition('<', key)
        return c

    def __ge__(self, key):
        c = self.add_custom_condition('>=', key)
        return c

    def __gt__(self, key):
        c = self.add_custom_condition('>', key)
        return c

    def __ne__(self, key):
        c = self.add_custom_condition('NE', key)
        return c



import sasoptpy as so
import pandas as pd


def test(cas_conn):

    m = so.Model(name='factory_planning_2', session=cas_conn)

    # Input data
    product_list = ['prod{}'.format(i) for i in range(1, 8)]
    product_data = pd.DataFrame([10, 6, 8, 4, 11, 9, 3],
                                columns=['profit'], index=product_list)
    demand_data = [
        [500, 1000, 300, 300,  800, 200, 100],
        [600,  500, 200,   0,  400, 300, 150],
        [300,  600,   0,   0,  500, 400, 100],
        [200,  300, 400, 500,  200,   0, 100],
        [0,    100, 500, 100, 1000, 300,   0],
        [500,  500, 100, 300, 1100, 500,  60]]
    demand_data = pd.DataFrame(
        demand_data, columns=product_list, index=range(1, 7))
    machine_type_product_data = [
        ['grinder', 0.5,  0.7,  0,    0,    0.3,  0.2, 0.5],
        ['vdrill',  0.1,  0.2,  0,    0.3,  0,    0.6, 0],
        ['hdrill',  0.2,  0,    0.8,  0,    0,    0,   0.6],
        ['borer',   0.05, 0.03, 0,    0.07, 0.1,  0,   0.08],
        ['planer',  0,    0,    0.01, 0,    0.05, 0,   0.05]]
    machine_type_product_data = \
        pd.DataFrame(machine_type_product_data, columns=['machine_type'] +
                     product_list).set_index(['machine_type'])
    machine_types_data = [
        ['grinder', 4, 2],
        ['vdrill',  2, 2],
        ['hdrill',  3, 3],
        ['borer',   1, 1],
        ['planer',  1, 1]]
    machine_types_data = pd.DataFrame(machine_types_data, columns=[
        'machine_type', 'num_machines', 'num_machines_needing_maintenance'])\
        .set_index(['machine_type'])

    store_ub = 100
    storage_cost_per_unit = 0.5
    final_storage = 50
    num_hours_per_period = 24 * 2 * 8

    # Problem definition
    PRODUCTS = product_list
    profit = product_data['profit']
    PERIODS = range(1, 7)
    MACHINE_TYPES = machine_types_data.index.tolist()

    num_machines = machine_types_data['num_machines']

    make = m.add_variables(PRODUCTS, PERIODS, lb=0, name='make')
    sell = m.add_variables(PRODUCTS, PERIODS, lb=0, ub=demand_data.transpose(),
                           name='sell')

    store = m.add_variables(PRODUCTS, PERIODS, lb=0, ub=store_ub, name='store')
    for p in PRODUCTS:
        store[p, 6].set_bounds(lb=final_storage, ub=final_storage)

    storageCost = so.expr_sum(
        storage_cost_per_unit * store[p, t] for p in PRODUCTS for t in PERIODS)
    revenue = so.expr_sum(profit[p] * sell[p, t]
                           for p in PRODUCTS for t in PERIODS)
    m.set_objective(revenue-storageCost, sense=so.MAX, name='total_profit')

    num_machines_needing_maintenance = \
        machine_types_data['num_machines_needing_maintenance']
    numMachinesDown = m.add_variables(MACHINE_TYPES, PERIODS, vartype=so.INT,
                                      lb=0, name='numMachinesDown')

    production_time = machine_type_product_data
    m.add_constraints((
        so.expr_sum(production_time.at[mc, p] * make[p, t] for p in PRODUCTS)
        <= num_hours_per_period *
        (num_machines[mc] - numMachinesDown[mc, t])
        for mc in MACHINE_TYPES for t in PERIODS), name='machine_hours_con')

    m.add_constraints((so.expr_sum(numMachinesDown[mc, t] for t in PERIODS) ==
                       num_machines_needing_maintenance[mc]
                       for mc in MACHINE_TYPES), name='maintenance_con')

    m.add_constraints(((store[p, t-1] if t-1 in PERIODS else 0) + make[p, t] ==
                      sell[p, t] + store[p, t]
                      for p in PRODUCTS for t in PERIODS),
                      name='flow_balance_con')

    res = m.solve()
    if res is not None:
        print(so.get_solution_table(make, sell, store))
        print(so.get_solution_table(numMachinesDown).unstack(level=-1))

    print(m.get_solution_summary())
    print(m.get_problem_summary())

    return m.get_objective_value()




from sasoptpy.abstract.util import *
from .parameter import Parameter, ParameterValue
from .parameter_group import ParameterGroup
from .set import Set, InlineSet
from .set_iterator import SetIterator, SetIteratorGroup
from sasoptpy.abstract.statement import *
from sasoptpy.abstract.implicit_variable import ImplicitVar
import sasoptpy.abstract.math
from .shadow import ShadowVariable, ShadowConstraint
from .range import AbstractRange
from .condition import Condition, Conditional



from .package_utils import load_package_globals
from .package_utils import *
from .user_utils import *
import sasoptpy


def to_expression(item):
    if hasattr(item, '_expr'):
        return item._expr()
    else:
        return sasoptpy.util.package_utils._to_sas_string(item)


def to_definition(item):
    return item._defn()


def to_optmodel(item, **kwargs):
    return sasoptpy.interface.to_optmodel(item, **kwargs)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for core classes.
"""

from collections import OrderedDict
from difflib import SequenceMatcher
import inspect
import os
import unittest
import warnings
from inspect import cleandoc

import sasoptpy as so
from tests.swat_config import create_cas_connection


class MockSASconfig:

    def __init__(self, name):
        self.name = name

class SASsession:

    def __init__(self, cfgname):
        import saspy
        self.sascfg = MockSASconfig(name=cfgname)

class TestModel(unittest.TestCase):
    """
    Unit tests for :class:`sasoptpy.Model` objects
    """

    @classmethod
    def setUpClass(cls):
        cls.conn = None
        from swat import CAS, SWATError
        try:
            cls.conn = create_cas_connection()
        except SWATError:
            warnings.warn('CAS connection is not available',
                          RuntimeWarning)
        except TypeError:
            warnings.warn('CAS variables are not available',
                          RuntimeWarning)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def setUp(self):
        pass

    @classmethod
    def get_standard_model(cls, name):
        m = so.Model(name=name)
        x = m.add_variable(name='x')
        y = m.add_variables(2, name='y')
        c1 = m.add_constraint(x <= 5, name='c1')
        c2 = m.add_constraints((y[i] <= 3 for i in range(2)), name='c2')
        return m

    def test_initialize(self):
        m = so.Model(name='test_initialize', session=None)
        self.assertEqual(type(m), so.Model)

    def test_comparison(self):
        model1 = so.Model(name='test_equal_1', session=None)
        model2 = so.Model(name='test_equal_2', session=None)
        self.assertFalse(model1 == model2)

        model3 = model1
        self.assertTrue(model1 == model3)

        def invalid_comparison():
            _ = model1 == list()
        self.assertWarns(RuntimeWarning, invalid_comparison)

    def test_get_name(self):
        m = so.Model(name='m')
        self.assertEqual(m.get_name(), 'm')

    def test_adding_variable(self):
        m = so.Model(name='test_add_variable')

        x = m.add_variable(name='x')
        y = m.add_variable(name='y', vartype=so.INT)
        z = m.add_variable(name='z', lb=1, ub=10)
        w = m.add_variable(name='w', init=5)
        u = so.Variable(name='u')
        m.include(u)
        self.assertEqual(m.get_variables(), [x, y, z, w, u])
        self.assertEqual(m.get_variable_dict(), {'x': x, 'y': y, 'z': z,
                                                 'w': w, 'u': u})
        self.assertIs(m.get_variable('x'), x)
        self.assertIs(m.get_variable('t'), None)

    def test_duplicate_variables(self):

        m = so.Model(name='test_duplicate_variables')

        def add_multi_var():
            x = m.add_variable(name='x', lb=2)
            x2 = m.add_variable(name='x', lb=1)

        self.assertWarns(UserWarning, add_multi_var)
        self.assertEqual(m.to_optmodel(), cleandoc("""
            proc optmodel;
               min test_duplicate_variables_obj = 0;
               var x >= 1;
               solve;
            quit;"""))

    def test_dropping_variable(self):
        m = so.Model(name='test_drop_variable')
        x = m.add_variable(name='x')
        self.assertIs(m.get_variables()[0], x)
        self.assertIs(m.get_variable_dict()['x'], x)
        m.drop_variable(x)
        self.assertEqual(m.get_variables(), [])
        self.assertEqual(m.get_variable_dict(), {})
        m.include(x)
        self.assertIs(m.get_variable_dict()['x'], x)
        m.drop(x)
        self.assertEqual(m.get_variable_dict(), {})

    def test_drop_restore_var(self):
        m = so.Model(name='test_drop_restore')
        x = m.add_variable(name='x')
        y = m.add_variables(5, name='y')
        m.set_objective(y[3], sense=so.minimize, name='obj')
        self.assertEqual(m.to_optmodel(), cleandoc('''
            proc optmodel;
               var x;
               var y {{0,1,2,3,4}};
               min obj = y[3];
               solve;
            quit;'''))
        m.drop_variable(x)
        m.drop_variable(y[1])
        m.drop_variable(y[2])
        self.assertEqual(m.to_optmodel(), cleandoc('''
            proc optmodel;
               var y {{0,1,2,3,4}};
               min obj = y[3];
               drop y[1] y[2];
               solve;
            quit;'''))
        m.restore_variable(x)
        m.restore_variable(y[2])
        self.assertEqual(m.to_optmodel(), cleandoc('''
            proc optmodel;
               var x;
               var y {{0,1,2,3,4}};
               min obj = y[3];
               drop y[1];
               solve;
            quit;'''))

    def test_adding_vargroup(self):
        m = so.Model(name='test_add_vg')

        x = m.add_variables(2, name='x')
        y = m.add_variables(['a', 'b'], name='y', vartype=so.BIN)
        I = so.abstract.Set(name='I')
        z = m.add_variables(I, name='z', lb=1, ub=10, init=5)
        w = so.VariableGroup(5, name='w')
        m.include(w)
        vars = [('x', x), ('y', y), ('z', z), ('w', w)]
        self.assertEqual(m.get_grouped_variables(), OrderedDict(vars))
        self.assertIs(m.get_variable('x')[0], x[0])

    def test_dropping_vargroup(self):
        m = so.Model(name='test_drop_vg')
        x = m.add_variables(2, name='x')
        self.assertEqual(m.get_grouped_variables(), OrderedDict([('x', x)]))
        m.drop_variables(x)
        self.assertEqual(m.get_grouped_variables(), OrderedDict())
        m.include(x)
        self.assertEqual(m.get_grouped_variables(), OrderedDict([('x', x)]))
        m.drop(x)
        self.assertEqual(m.get_grouped_variables(), OrderedDict())

    def test_adding_constraint(self):
        m = so.Model(name='test_add_constraint')
        x = m.add_variable(name='x')

        c1 = m.add_constraint(x <= 5, name='c1')
        c2 = m.add_constraint(2 * x + x ** 5 >= 1, name='c2')
        self.assertEqual([c1, c2], m.get_constraints())
        self.assertEqual({'c1': c1, 'c2': c2}, m.get_constraints_dict())

        def invalid_constraint():
            from math import inf
            c3 = m.add_constraint(x <= inf, name='c3')
        self.assertRaises(ValueError, invalid_constraint)

        cx = m.get_constraint('c1')
        self.assertEqual(cx, c1)
        cy = m.get_constraint('c3')
        self.assertEqual(cy, None)

    def test_duplicate_constraints(self):

        m = so.Model(name='test_duplicate_constraints')

        def add_multi_con():
            x = m.add_variable(name='x')

            c1 = m.add_constraint(x <= 5, name='c')
            c2 = m.add_constraint(x <= 5, name='c')

        self.assertWarns(UserWarning, add_multi_con)
        self.assertEqual(m.to_optmodel(), cleandoc("""
            proc optmodel;
               min test_duplicate_constraints_obj = 0;
               var x;
               con c : x <= 5;
               solve;
            quit;"""))

    def test_drop_restore_cons(self):
        m = so.Model(name='test_drop_restore_constraints')
        x = m.add_variable(name='x')
        y = m.add_variables(5, name='y')
        m.set_objective(y[3], sense=so.minimize, name='obj')

        c1 = m.add_constraint(x <= 5, name='c1')
        c2 = m.add_constraints((y[i] <= i for i in range(5)), name='c2')
        self.assertEqual(m.to_optmodel(), cleandoc("""
            proc optmodel;
               var x;
               var y {{0,1,2,3,4}};
               min obj = y[3];
               con c1 : x <= 5;
               con c2_0 : y[0] <= 0;
               con c2_1 : y[1] <= 1;
               con c2_2 : y[2] <= 2;
               con c2_3 : y[3] <= 3;
               con c2_4 : y[4] <= 4;
               solve;
            quit;"""))
        m.drop_constraint(c1)
        m.drop_constraint(c2[1])
        m.drop_constraint(c2[2])
        self.assertEqual(m.to_optmodel(), cleandoc("""
            proc optmodel;
               var x;
               var y {{0,1,2,3,4}};
               min obj = y[3];
               con c2_0 : y[0] <= 0;
               con c2_1 : y[1] <= 1;
               con c2_2 : y[2] <= 2;
               con c2_3 : y[3] <= 3;
               con c2_4 : y[4] <= 4;
               drop c2_1 c2_2;
               solve;
            quit;"""))
        m.restore_constraint(c1)
        m.restore_constraint(c2[2])
        self.assertEqual(m.to_optmodel(), cleandoc("""
            proc optmodel;
               var x;
               var y {{0,1,2,3,4}};
               min obj = y[3];
               con c1 : x <= 5;
               con c2_0 : y[0] <= 0;
               con c2_1 : y[1] <= 1;
               con c2_2 : y[2] <= 2;
               con c2_3 : y[3] <= 3;
               con c2_4 : y[4] <= 4;
               drop c2_1;
               solve;
            quit;"""))

    def test_dropping_constraint(self):
        m = so.Model(name='test_drop_constraint')
        x = m.add_variable(name='x')
        c1 = m.add_constraint(x <= 5, name='c1')
        self.assertEqual({'c1': c1}, m.get_constraints_dict())
        m.drop_constraint(c1)
        self.assertEqual({}, m.get_constraints_dict())
        m.include(c1)
        self.assertEqual({'c1': c1}, m.get_constraints_dict())
        m.drop(c1)
        self.assertEqual({}, m.get_constraints_dict())

    def test_adding_constraints(self):
        m = so.Model(name='test_add_cg')
        x = m.add_variables(5, name='x')

        c1 = m.add_constraints((x[i] >= i for i in range(5)), name='c1')
        self.assertEqual(OrderedDict([('c1', c1)]), m.get_grouped_constraints())
        self.assertEqual(c1, m.get_constraint('c1'))

        c2 = so.ConstraintGroup((i * x[i] <= 10 for i in range(5)), name='c2')
        m.include(c2)
        grouped_con_dict = OrderedDict([('c1', c1), ('c2', c2)])
        self.assertEqual(grouped_con_dict, m.get_grouped_constraints())

        def warn_user_single_constraint():
            c3 = m.add_constraints(x[0] >= 1, name='c3')
        self.assertWarns(UserWarning, warn_user_single_constraint)

    def test_dropping_constraints(self):
        m = so.Model(name='test_drop_cg')
        x = m.add_variables(2, name='x')
        c1 = m.add_constraints((x[i] <= i for i in range(2)), name='c1')
        self.assertEqual(m.get_grouped_constraints(), OrderedDict([('c1', c1)]))
        m.drop_constraints(c1)
        self.assertEqual(m.get_grouped_constraints(), OrderedDict())
        m.include(c1)
        self.assertEqual(m.get_grouped_constraints(), OrderedDict([('c1', c1)]))
        m.drop(c1)
        self.assertEqual(m.get_grouped_constraints(), OrderedDict())

    def test_add_set(self):
        m = so.Model(name='test_add_set')
        I = m.add_set(name='I', init=2)
        self.assertEqual(m.get_sets(), [I])
        self.assertEqual(so.to_definition(m.get_sets()[0]), "set I init 2;")

    def test_add_parameter(self):
        m = so.Model(name='test_add_parameter')
        p = m.add_parameter(name='p', init=10)
        I = m.add_set(name='I')
        r = m.add_parameter(I, name='r', init=5)
        self.assertEqual([p, r], m.get_parameters())
        m.drop(r)
        self.assertEqual([p], m.get_parameters())

    def test_add_implicit_var(self):
        m = so.Model(name='test_add_impvar')
        x = m.add_variables(5, name='x')
        y = m.add_implicit_variable((i * x[i] + x[i] ** 2 for i in range(5)),
                                    name='y')
        self.assertEqual([y], m.get_implicit_variables())

    def test_add_literal_statement(self):
        m = so.Model(name='test_add_literal_statement')
        m.set_objective(0, name='empty_obj')
        m.add_statement('var x {0,1};')
        m.add_statement('solve;')
        self.assertEqual(
            m.to_optmodel(solve=False),
            inspect.cleandoc('''
            proc optmodel;
               min empty_obj = 0;
               var x {0,1};
               solve;
            quit;'''))
        s = so.abstract.LiteralStatement('print x;')
        m.include(s)
        self.assertEqual(
            m.to_optmodel(solve=False),
            inspect.cleandoc('''
            proc optmodel;
               min empty_obj = 0;
               var x {0,1};
               solve;
               print x;
            quit;'''))
        m.drop(s)
        self.assertEqual(
            m.to_optmodel(solve=False),
            inspect.cleandoc('''
            proc optmodel;
               min empty_obj = 0;
               var x {0,1};
               solve;
            quit;'''))


    def test_add_abstract_statement(self):
        m = so.Model(name='m')
        x = m.add_variable(name='x')
        m.set_objective(x ** 2, sense=so.MIN, name='obj')
        s = so.abstract.LiteralStatement('expand;')
        m.add_statement(s)
        self.assertEqual(so.to_optmodel(m), inspect.cleandoc("""
            proc optmodel;
               var x;
               min obj = (x) ^ (2);
               expand;
               solve;
            quit;
            """))

    def test_postsolve_statement(self):
        m = so.Model(name='test_postsolve_statement')
        x = m.add_variable(name='x')
        c1 = m.add_constraint(x <= 10, name='c1')
        self.assertEqual(m.to_optmodel(), inspect.cleandoc("""
            proc optmodel;
               min test_postsolve_statement_obj = 0;
               var x;
               con c1 : x <= 10;
               solve;
            quit;"""))

        m.add_postsolve_statement('print x;')
        self.assertEqual(m.to_optmodel(), inspect.cleandoc("""
        proc optmodel;
           min test_postsolve_statement_obj = 0;
           var x;
           con c1 : x <= 10;
           solve;
           print x;
        quit;"""))

        m.add_postsolve_statement(so.abstract.LiteralStatement('expand;'))
        self.assertEqual(m.to_optmodel(), inspect.cleandoc("""
        proc optmodel;
           min test_postsolve_statement_obj = 0;
           var x;
           con c1 : x <= 10;
           solve;
           print x;
           expand;
        quit;"""))

    def test_include_model(self):
        m1 = so.Model(name='test_copy_model_1')
        x = m1.add_variable(name='x')
        y = m1.add_variables(2, name='y')
        c1 = m1.add_constraint(x + y[0] >= 2, name='c1')
        c2 = m1.add_constraints((x - y[i] <= 10 for i in range(2)), name='c2')
        m1.set_objective(2 * x + y[0] + 3 * y[1], name='model_obj')

        m2 = so.Model(name='test_copy_model_2')
        m2.include(m1)
        vars = OrderedDict([('x', x), ('y', y)])
        self.assertEqual(m2.get_grouped_variables(), vars)
        cons = OrderedDict([('c1', c1), ('c2', c2)])
        self.assertEqual(m2.get_grouped_constraints(), cons)
        self.assertEqual(m2.to_optmodel(),inspect.cleandoc("""
            proc optmodel;
               var x;
               var y {{0,1}};
               con c1 : x + y[0] >= 2;
               con c2_0 : x - y[0] <= 10;
               con c2_1 : x - y[1] <= 10;
               min model_obj = 2 * x + y[0] + 3 * y[1];
               solve;
            quit;"""))

    def test_set_get_objective(self):
        m = so.Model(name='test_set_get_objective')
        x = m.add_variable(name='x')

        # Regular objective
        obj1 = m.set_objective(2 * x, sense=so.MIN, name='obj1')
        self.assertIs(obj1, m.get_objective())

        # Multi objective
        obj2 = m.set_objective(5 * x, sense=so.MIN, name='obj2')
        self.assertIs(obj2, m.get_objective())
        obj3 = m.append_objective(10 * x, sense=so.MIN, name='obj3')
        self.assertEqual([obj2, obj3], m.get_all_objectives())
        self.assertEqual(
            m.to_optmodel(),
            inspect.cleandoc("""
            proc optmodel;
               var x;
               min obj2 = 5 * x;
               min obj3 = 10 * x;
               solve;
            quit;"""))

    def test_get_objective_value(self):
        m = so.Model(name='test_objective_value')
        x = m.add_variable(name='x')
        m.set_objective(x ** 2 - 4 * x + 5, sense=so.MIN, name='nonlinear')
        x.set_value(3)
        self.assertEqual(m.get_objective_value(), 2)

        if TestModel.conn:
            m.set_session(TestModel.conn)
            m.solve()
            self.assertEqual(m.get_objective_value(), 1)
            self.assertEqual(x.get_value(), 2)
        else:
            self.skipTest('No CAS connection available, skipping ' +
                              'objective value test')

        def zero_div_error():
            m.set_objective(x / x, sense=so.MIN, name='nonlinear2')
            x.set_value(0)
            m.clear_solution()
            m.get_objective_value()
        self.assertRaises(ZeroDivisionError, zero_div_error)

    def test_variable_coef(self):
        m = so.Model(name='test_get_variable_coef')
        x = m.add_variable(name='x')
        m.set_objective(5 * x, sense=so.MIN, name='obj1')

        self.assertEqual(m.get_variable_coef(x), 5)
        self.assertEqual(m.get_variable_coef('x'), 5)

        y = so.Variable(name='y')
        def variable_not_in_model():
            return m.get_variable_coef(y)
        self.assertRaises(RuntimeError, variable_not_in_model)

        m.set_objective(2 * x + y ** 2, sense=so.MIN, name='obj1')
        self.assertEqual(m.get_variable_coef('x'), 2)
        def nonlinear_objective():
            return m.get_variable_coef('y')
        self.assertWarns(RuntimeWarning, nonlinear_objective)

    def test_get_variable_value(self):

        if TestModel.conn is None:
            self.skipTest('Session is not available')

        m = so.Model(name='test_get_var_value')
        x = m.add_variable(name='x', lb=1.5, ub=10, vartype=so.INT)
        m.set_objective(x, sense=so.MIN, name='obj1')
        m.set_session(TestModel.conn)
        m.solve(verbose=True)
        self.assertEqual(m.get_variable_value(x), 2)

        I = m.add_set(name='I', value=range(2))
        y = m.add_variables(I, name='y', lb=0.5)
        m.set_objective(x + y[0] + y[1], sense=so.MIN, name='obj1')
        m.solve()
        self.assertEqual(m.get_variable_value(y[0]), 0.5)
        def get_variable_warning():
            self.assertEqual(m.get_variable_value('z'), None)
        self.assertWarns(UserWarning, get_variable_warning)

        m2 = so.Model(name='test_get_var_value_copy')
        m2.include(m)
        z = so.Variable(name='z')
        def raise_solution_error():
            return m2.get_variable_value(z)
        self.assertRaises(RuntimeError, raise_solution_error)

        m.add_variable(name='var with invalid name')
        def raise_syntax_error():
            return m.solve()
        self.assertRaises(SyntaxError, raise_syntax_error)

    def test_get_variable_value_abstract(self):
        if TestModel.conn is None:
            self.skipTest('Session is not available')

        import pandas as pd
        so.reset()

        m = so.Model(name='abstract_model')
        df = pd.DataFrame([
            ['a', 1],
            ['b', 2]
        ], columns=['tag', 'val'])
        idx = so.Set(name='idx', settype=so.STR)
        varlb = so.ParameterGroup(idx, name='varlb')
        m.include(idx, varlb)

        table = TestModel.conn.upload_frame(df, casout='server_data')
        from sasoptpy.actions import read_data
        r = read_data(
            table=table,
            index={'target': idx, 'key': 'tag'},
            columns=[
                {'target': varlb, 'column': 'val'}
            ]
        )
        m.include(r)
        y = so.VariableGroup(idx, name='y')
        c = so.ConstraintGroup((y[i] >= varlb[i] for i in idx), name='c')
        m.include(y, c)
        self.assertEqual(m.to_optmodel(), inspect.cleandoc("""
            proc optmodel;
               min abstract_model_obj = 0;
               set <str> idx;
               num varlb {idx};
               read data SERVER_DATA into idx=[tag] varlb=val;
               var y {{idx}};
               con c {o8 in idx} : y[o8] - varlb[o8] >= 0;
               solve;
            quit;
            """))
        m.set_session(TestModel.conn)
        m.solve()
        self.assertEqual(m.get_variable_value(y['a']), 1)
        self.assertEqual(m.get_statements(), [r])


    def test_get_summaries(self):
        if not TestModel.conn:
            self.skipTest('Session is not available')
        m = so.Model(name='test_get_summaries', session=TestModel.conn)
        x = m.add_variable(name='x', lb=1)
        y = m.add_variables(2, name='y', lb=1)
        m.set_objective(x + y[0], sense=so.MIN, name='obj1')
        m.add_constraint(x + 2 *y[0] + 3*y[1] >= 10, name='con1')
        m.solve()
        self.assertEqual(m.get_problem_summary().to_string(),
                         inspect.cleandoc("""
                            Selected Rows from Table PROBLEMSUMMARY

                                                            Value
                            Label                                
                            Objective Sense          Minimization
                            Objective Function               obj1
                            Objective Type                 Linear
                                                                 
                            Number of Variables                 3
                            Bounded Above                       0
                            Bounded Below                       3
                            Bounded Below and Above             0
                            Free                                0
                            Fixed                               0
                                                                 
                            Number of Constraints               1
                            Linear LE (<=)                      0
                            Linear EQ (=)                       0
                            Linear GE (>=)                      1
                            Linear Range                        0
                                                                 
                            Constraint Coefficients             3"""))

        seq = SequenceMatcher(None, m.get_solution_summary().to_string(),
                              inspect.cleandoc(
            """
            Selected Rows from Table SOLUTIONSUMMARY

                                         Value
            Label                             
            Solver                          LP
            Algorithm             Dual Simplex
            Objective Function            obj1
            Solution Status            Optimal
            Objective Value                  2
                                              
            Primal Infeasibility             0
            Dual Infeasibility               0
            Bound Infeasibility              0
                                              
            Iterations                       0
            Presolve Time                 0.00
            Solution Time                 0.00"""
        ))
        # There is a chance that the solution time is slightly different
        self.assertTrue(seq.ratio() > 0.99)

    def test_get_solution(self):
        if not TestModel.conn:
            self.skipTest('No session is defined, skipping get solution test')
        import pandas as pd
        m = so.Model(name='test_get_soln', session=TestModel.conn)
        data = [
            ['pen', 1, 3, 11],
            ['mug', 15, 10, 5],
            ['watch', 50, 2, 2],
            ['pc', 1500, 200, 1]
        ]
        data = pd.DataFrame(data, columns=['item', 'value', 'weight', 'ub'])
        data = data.set_index(['item'])
        items = data.index
        get = m.add_variables(items, name='get', vartype=so.INT, lb=0)
        value = data['value']
        weight = data['weight']
        ub = data['ub']
        m.set_objective(so.expr_sum(get[i] * value[i] for i in items),
                        sense=so.MAX, name='obj1')
        m.add_constraint(so.expr_sum(get[i] * weight[i] for i in items)
                         <= 210, name='value_total')
        m.add_constraints((get[i] <= ub[i] for i in items), name='upper_bound')

        # Regular solve and regular get
        m.solve(verbose=True)
        self.assertEqual(m.get_solution().to_string(), inspect.cleandoc(
            """
            Selected Rows from Table SOLUTION

                 i         var  value   lb             ub  rc
            0  1.0    get[pen]    2.0 -0.0  1.797693e+308 NaN
            1  2.0    get[mug]   -0.0 -0.0  1.797693e+308 NaN
            2  3.0  get[watch]    2.0 -0.0  1.797693e+308 NaN
            3  4.0     get[pc]    1.0 -0.0  1.797693e+308 NaN
            """
        ))

        self.assertEqual(m.get_solution(vtype='dual').to_string(),
                         inspect.cleandoc(
            """
            Selected Rows from Table DUAL

                 j                con  value  dual
            0  1.0        value_total  210.0   NaN
            1  2.0    upper_bound_pen    2.0   NaN
            2  3.0    upper_bound_mug   -0.0   NaN
            3  4.0  upper_bound_watch    2.0   NaN
            4  5.0     upper_bound_pc    1.0   NaN
            """
        ))

        m.solve(mps=True, options={'maxpoolsols': 3}, verbose=True)
        self.assertEqual(m.get_solution().to_string(), inspect.cleandoc(
            """
            Selected Rows from Table PRIMAL

                       var   lb             ub  value  solution
            0     get[pen]  0.0  1.797693e+308    2.0       1.0
            1     get[mug]  0.0  1.797693e+308    0.0       1.0
            2   get[watch]  0.0  1.797693e+308    2.0       1.0
            3      get[pc]  0.0  1.797693e+308    1.0       1.0
            4     get[pen]  0.0  1.797693e+308    1.0       2.0
            5     get[mug]  0.0  1.797693e+308    0.0       2.0
            6   get[watch]  0.0  1.797693e+308    1.0       2.0
            7      get[pc]  0.0  1.797693e+308    1.0       2.0
            8     get[pen]  0.0  1.797693e+308    0.0       3.0
            9     get[mug]  0.0  1.797693e+308    0.0       3.0
            10  get[watch]  0.0  1.797693e+308    0.0       3.0
            11     get[pc]  0.0  1.797693e+308    0.0       3.0
            """
        ))
        self.assertEqual(m.get_solution('dual').to_string(), inspect.cleandoc(
            """
            Selected Rows from Table DUAL

                                 con  value  solution
            0            value_total  210.0       1.0
            1     upper_bound['pen']    2.0       1.0
            2     upper_bound['mug']    0.0       1.0
            3   upper_bound['watch']    2.0       1.0
            4      upper_bound['pc']    1.0       1.0
            5            value_total  205.0       2.0
            6     upper_bound['pen']    1.0       2.0
            7     upper_bound['mug']    0.0       2.0
            8   upper_bound['watch']    1.0       2.0
            9      upper_bound['pc']    1.0       2.0
            10           value_total    0.0       3.0
            11    upper_bound['pen']    0.0       3.0
            12    upper_bound['mug']    0.0       3.0
            13  upper_bound['watch']    0.0       3.0
            14     upper_bound['pc']    0.0       3.0
            """
        ))
        self.assertEqual(m.get_solution(pivot=True).to_string(),
                         inspect.cleandoc(
            """
            Selected Rows from Table PRIMAL

            Fetch       1.0  2.0  3.0
            Fetch                    
            get[mug]    0.0  0.0  0.0
            get[pc]     1.0  1.0  0.0
            get[pen]    2.0  1.0  0.0
            get[watch]  2.0  1.0  0.0
            """
        ))
        self.assertEqual(m.get_solution('dual', pivot=True).to_string(),
                         inspect.cleandoc(
            """
            Selected Rows from Table DUAL

            Fetch                   1.0    2.0  3.0
            Fetch                                  
            upper_bound['mug']      0.0    0.0  0.0
            upper_bound['pc']       1.0    1.0  0.0
            upper_bound['pen']      2.0    1.0  0.0
            upper_bound['watch']    2.0    1.0  0.0
            value_total           210.0  205.0  0.0
            """
        ))
        self.assertEqual(m.get_solution('primal', solution=2).to_string(),
                         inspect.cleandoc(
            """
            Selected Rows from Table PRIMAL

                      var   lb             ub  value  solution
            4    get[pen]  0.0  1.797693e+308    1.0       2.0
            5    get[mug]  0.0  1.797693e+308    0.0       2.0
            6  get[watch]  0.0  1.797693e+308    1.0       2.0
            7     get[pc]  0.0  1.797693e+308    1.0       2.0
            """
                         ))
        self.assertEqual(m.get_solution('dual', solution=3).to_string(),
                         inspect.cleandoc(
            """
            Selected Rows from Table DUAL

                                 con  value  solution
            10           value_total    0.0       3.0
            11    upper_bound['pen']    0.0       3.0
            12    upper_bound['mug']    0.0       3.0
            13  upper_bound['watch']    0.0       3.0
            14     upper_bound['pc']    0.0       3.0
            """
                         ))
        m.print_solution()

        def third_type():
            m.get_solution('x')
        self.assertRaises(ValueError, third_type)

    def test_set_coef(self):
        m = so.Model(name='test_set_coef')
        x = m.add_variable(name='x')
        y = m.add_variables(2, name='y')
        z = m.add_variable(name='z')
        obj = m.set_objective(2*x + 3*y[0] + 2*y[1], name='obj', sense=so.MIN)
        c1 = m.add_constraint(2* x + 5 * y[0] + 7 * y[1] <= 15, name='c1')
        self.assertEqual(m.get_variable_coef(x), 2)
        m.set_variable_coef(x, 3)
        self.assertEqual(m.get_variable_coef(x), 3)
        self.assertEqual(m.get_variable_coef(z), 0)
        m.set_variable_coef(z, 1)
        self.assertEqual(m.get_variable_coef(z), 1)

    def test_to_mps(self):
        m = so.Model(name='test_to_mps')
        x = m.add_variable(name='x', lb=0, ub=5, vartype=so.INT)
        y = m.add_variables(2, name='y', lb=1)
        m.set_objective(x + y[0], sense=so.MIN, name='xyobj')
        self.assertEqual(m.to_mps().to_string(), inspect.cleandoc(
            """
                 Field1    Field2       Field3  Field4    Field5  Field6  _id_
            0      NAME            test_to_mps     0.0               0.0     1
            1      ROWS                            NaN               NaN     2
            2       MIN     xyobj                  NaN               NaN     3
            3   COLUMNS                            NaN               NaN     4
            4            MARK0000     'MARKER'     NaN  'INTORG'     NaN     5
            5                   x        xyobj     1.0               NaN     6
            6            MARK0001     'MARKER'     NaN  'INTEND'     NaN     7
            7                y[0]        xyobj     1.0               NaN     8
            8                y[1]        xyobj     0.0               NaN     9
            9       RHS                            NaN               NaN    10
            10   RANGES                            NaN               NaN    11
            11   BOUNDS                            NaN               NaN    12
            12       LO       BND            x     0.0               NaN    13
            13       UP       BND            x     5.0               NaN    14
            14       LO       BND         y[0]     1.0               NaN    15
            15       LO       BND         y[1]     1.0               NaN    16
            16   ENDATA                            0.0               0.0    17
            """
        ))
        m.set_objective(x + 10, name='o', sense=so.MAX)
        self.assertEqual(m.to_mps(constant=True).to_string(),
                         inspect.cleandoc(
        """
             Field1        Field2        Field3  Field4    Field5  Field6  _id_
        0      NAME                 test_to_mps     0.0               0.0     1
        1      ROWS                                 NaN               NaN     2
        2       MAX    o_constant                   NaN               NaN     3
        3   COLUMNS                                 NaN               NaN     4
        4                MARK0000      'MARKER'     NaN  'INTORG'     NaN     5
        5                       x    o_constant     1.0               NaN     6
        6                MARK0001      'MARKER'     NaN  'INTEND'     NaN     7
        7                    y[0]    o_constant     0.0               NaN     8
        8                    y[1]    o_constant     0.0               NaN     9
        9            obj_constant    o_constant     1.0               NaN    10
        10      RHS                                 NaN               NaN    11
        11   RANGES                                 NaN               NaN    12
        12   BOUNDS                                 NaN               NaN    13
        13       LO           BND             x     0.0               NaN    14
        14       UP           BND             x     5.0               NaN    15
        15       LO           BND          y[0]     1.0               NaN    16
        16       LO           BND          y[1]     1.0               NaN    17
        17       FX           BND  obj_constant    10.0               NaN    18
        18   ENDATA                                 0.0               0.0    19
        """
                         ))

        # Add invalid constraints for the frame
        c1 = m.add_constraint(y[0] + x >= 0, name='zero_lb')
        c2 = m.add_constraint(y[0] <= 100, name='inf_ub')
        from math import inf
        c2.set_rhs(inf)
        self.assertEqual(m.to_mps().to_string(), inspect.cleandoc(
        """
             Field1        Field2        Field3  Field4    Field5  Field6  _id_
        0      NAME                 test_to_mps     0.0               0.0     1
        1      ROWS                                 NaN               NaN     2
        2       MAX    o_constant                   NaN               NaN     3
        3         G       zero_lb                   NaN               NaN     4
        4         L        inf_ub                   NaN               NaN     5
        5   COLUMNS                                 NaN               NaN     6
        6                MARK0000      'MARKER'     NaN  'INTORG'     NaN     7
        7                       x    o_constant     1.0   zero_lb     1.0     8
        8                MARK0001      'MARKER'     NaN  'INTEND'     NaN     9
        9                    y[0]       zero_lb     1.0    inf_ub     1.0    10
        10                   y[1]    o_constant     0.0               NaN    11
        11           obj_constant    o_constant     1.0               NaN    12
        12      RHS                                 NaN               NaN    13
        13   RANGES                                 NaN               NaN    14
        14   BOUNDS                                 NaN               NaN    15
        15       LO           BND             x     0.0               NaN    16
        16       UP           BND             x     5.0               NaN    17
        17       LO           BND          y[0]     1.0               NaN    18
        18       LO           BND          y[1]     1.0               NaN    19
        19       FX           BND  obj_constant    10.0               NaN    20
        20   ENDATA                                 0.0               0.0    21
        """
        ))

        u = m.add_variable(name='u')
        t = m.add_variable(name='t', vartype=so.BIN)
        m.drop_constraints(c1, c2)
        m.add_constraint(x + 2*y[0] == [3, 8], name='range_con')
        self.assertEqual(m.to_mps().to_string(), inspect.cleandoc(
        """
             Field1        Field2        Field3  Field4     Field5  Field6  _id_
        0      NAME                 test_to_mps     0.0                0.0     1
        1      ROWS                                 NaN                NaN     2
        2       MAX    o_constant                   NaN                NaN     3
        3         E     range_con                   NaN                NaN     4
        4   COLUMNS                                 NaN                NaN     5
        5                MARK0000      'MARKER'     NaN   'INTORG'     NaN     6
        6                       x    o_constant     1.0  range_con     1.0     7
        7                MARK0001      'MARKER'     NaN   'INTEND'     NaN     8
        8                    y[0]     range_con     2.0                NaN     9
        9                    y[1]    o_constant     0.0                NaN    10
        10           obj_constant    o_constant     1.0                NaN    11
        11                      u    o_constant     0.0                NaN    12
        12                      t    o_constant     0.0                NaN    13
        13      RHS                                 NaN                NaN    14
        14                    RHS     range_con     3.0                NaN    15
        15   RANGES                                 NaN                NaN    16
        16                    rng     range_con     5.0                NaN    17
        17   BOUNDS                                 NaN                NaN    18
        18       LO           BND             x     0.0                NaN    19
        19       UP           BND             x     5.0                NaN    20
        20       LO           BND          y[0]     1.0                NaN    21
        21       LO           BND          y[1]     1.0                NaN    22
        22       FX           BND  obj_constant    10.0                NaN    23
        23       FR           BND             u     NaN                NaN    24
        24       BV           BND             t     1.0                NaN    25
        25   ENDATA                                 0.0                0.0    26
        """
        ))

        def get_frame_warning():
            r = m.to_frame()
        self.assertWarns(DeprecationWarning, get_frame_warning)

    def test_to_optmodel(self):
        m = so.Model(name='test_to_optmodel')
        self.assertEqual(m.to_optmodel(), inspect.cleandoc(
            """
            proc optmodel;
               min test_to_optmodel_obj = 0;
               solve;
            quit;
            """
        ))
        x = m.add_variable(name='x', init=5)
        e1 = m.set_objective(x, sense=so.MIN, name='e1')
        e2 = m.append_objective(x**2, sense=so.MAX, name='e2')
        response = m.to_optmodel(options={
            'with': 'blackbox',
            'relaxint': True,
            'obj': (e1, e2),
            'primalin': True,
        }, ods=True, primalin=True, parse=False)
        self.assertEqual(response, inspect.cleandoc(
            """
            proc optmodel;
               var x init 5;
               min e1 = x;
               max e2 = (x) ^ (2);
               solve with blackbox relaxint obj (e1 e2) / primalin;
               ods output PrintTable=primal_out;
               ods output PrintTable=dual_out;
               create data allsols from [s]=(1.._NVAR_) name=_VAR_[s].name {j in 1.._NSOL_} <col('sol_'||j)=_VAR_[s].sol[j]>;
            quit;
            """
        ))

        response = m.to_optmodel(options={
            'with': 'nlp',
            'multistart': {'loglevel': 3, 'maxstarts': 30}
        })
        self.assertEqual(response, inspect.cleandoc(
            """
            proc optmodel;
               var x init 5;
               min e1 = x;
               max e2 = (x) ^ (2);
               solve with nlp / multistart=(loglevel=3,maxstarts=30);
            quit;
            """
        ))

    def test_str(self):
        m = TestModel.get_standard_model(name='test_model_str')

        response = str(m)
        self.assertEqual(response, inspect.cleandoc(
            """
            Model: [
              Name: test_model_str
              Objective: MIN [0]
              Variables (3): [
                x
                y[0]
                y[1]
              ]
              Constraints (3): [
                x <=  5
                y[0] <=  3
                y[1] <=  3
              ]
            ]
            """
        ))
        if TestModel.conn:
            m.set_session(TestModel.conn)
            response = str(m)
            self.assertEqual(response, inspect.cleandoc(
                """
                Model: [
                  Name: test_model_str
                  Session: {}:{}
                  Objective: MIN [0]
                  Variables (3): [
                    x
                    y[0]
                    y[1]
                  ]
                  Constraints (3): [
                    x <=  5
                    y[0] <=  3
                    y[1] <=  3
                  ]
                ]
                """.format(os.environ.get('CASHOST'), os.environ.get('CASPORT'))
            ))

    def test_model_repr(self):
        m = so.Model(name='test_model_repr')
        self.assertEqual(repr(m), "sasoptpy.Model(name='test_model_repr')")
        s = SASsession(cfgname='winlocal')
        m.set_session(s)
        self.assertEqual(
            repr(m),
            "sasoptpy.Model(name='test_model_repr', "
            "session=saspy.SASsession(cfgname='winlocal'))")

        if TestModel.conn:
            m.set_session(TestModel.conn)
            cas_repr = repr(m.get_session())
            self.assertEqual(
                repr(m), "sasoptpy.Model(name='test_model_repr', session=" +
                cas_repr + ')')

        def invalid_session_type():
            w = 5
            m.set_session(w)
            rp = repr(m)
        self.assertRaises(TypeError, invalid_session_type)

    def test_defn(self):
        m = TestModel.get_standard_model('test_model_defn')
        self.assertEqual(so.to_definition(m), "problem test_model_defn "
                                              "include x y c1 c2;")

    def test_expr(self):
        m = TestModel.get_standard_model('test_model_expr')
        self.assertEqual(m.to_optmodel(), so.to_expression(m))

    def test_is_linear(self):
        m = TestModel.get_standard_model('test_model_linearity')
        self.assertEqual(so.is_linear(m), True)
        x = m.get_variable('x')
        qbound = m.add_constraint(x ** 2 + x <= 10, name='qbound')
        self.assertEqual(so.is_linear(m), False)
        m.drop_constraint(qbound)
        self.assertEqual(so.is_linear(m), True)
        m.set_objective(x ** 2, sense=so.MIN, name='x_squared')
        self.assertEqual(so.is_linear(m), False)

    def test_session_type(self):
        m = TestModel.get_standard_model('test_model_session_type')
        self.assertEqual(m.get_session_type(), None)
        if TestModel.conn:
            m.set_session(TestModel.conn)
            self.assertEqual(m.get_session_type(), 'CAS')

    def test_ub_set(self):
        m = so.Model(name='test_model_var_ub')
        x = m.add_variable(name='x')
        self.assertEqual(so.to_optmodel(m), cleandoc('''
            proc optmodel;
               min test_model_var_ub_obj = 0;
               var x;
               solve;
            quit;'''))
        x.set_bounds(ub=5)
        self.assertEqual(so.to_optmodel(m), cleandoc('''
            proc optmodel;
               min test_model_var_ub_obj = 0;
               var x <= 5;
               solve;
            quit;'''))

    def test_model_add(self):
        m = so.Model(name='test_add')
        x = so.Variable(name='x')
        self.assertEqual(m.get_variables(), [])
        m.add(x)
        self.assertEqual(m.get_variables(), [x])

    def test_model_session(self):
        m = so.Model(name='m')
        s = m.get_session()
        self.assertEqual(s, None)
        if TestModel.conn:
            m.set_session(TestModel.conn)
            self.assertEqual(m.get_session(), TestModel.conn)
            self.assertEqual(m.get_session_type(), 'CAS')

    def test_names(self):
        if TestModel.conn is None:
            self.skipTest('Session is not available')

        m = so.Model(name='test_var_names', session=TestModel.conn)
        a = ['apple', 'apple juice']
        x = m.add_variables(a, name='amount', lb=1)
        m.set_objective(so.expr_sum(x[i] for i in a), name='obj', sense=so.minimize)

        m.solve()
        for i in a:
            self.assertEqual(x[i].get_value(), 1.0)

    def test_export(self):
        m = TestModel.get_standard_model('test_model_export')
        x = m.get_variable('x')
        mps_text = m.export_mps(fetch=True)
        print(mps_text)
        self.assertEqual(mps_text.replace(' ', ''), inspect.cleandoc(
            """
            NAME     test_model_export
            ROWS
             MIN     test_model_export_obj
             L       c1
             L       c2[0]
             L       c2[1]
            COLUMNS
                     x                      c1     1.0
                     y[0]                   c2[0]  1.0
                     y[1]                   c2[1]  1.0
            RHS
                     RHS                    c1     5.0  c2[0]  3.0
                     RHS                    c2[1]  3.0
            RANGES
            BOUNDS
             FR      BND                    x
             FR      BND                    y[0]
             FR      BND                    y[1]
            ENDATA"""
        ).replace(' ', ''))
        
        m.add_constraint(x ** 2 + x <= 10, name='qb')
        def generate_error():
            m.export_mps()
        self.assertRaises(ValueError, generate_error)


    def tearDown(self):
        so.reset()





#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for core classes.
"""

import os
import sys
import unittest
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '..')))
from util import assert_equal_wo_temps


class TestConstraintGroup(unittest.TestCase):
    """
    Unit tests for :class:`sasoptpy.ConstraintGroup` objects
    """

    def setUp(self):
        pass

    def test_constructor(self):
        I = [1, 2, 3]
        x = so.VariableGroup(I, name='x')
        cg = so.ConstraintGroup((x[i] <= 3 for i in I), name='bound_con')
        self.assertEqual(
            repr(cg),
            "sasoptpy.ConstraintGroup([x[1] <=  3, x[2] <=  3, x[3] <=  3], name='bound_con')")

    def test_name(self):
        x = so.VariableGroup(2, name='x')
        c1 = so.ConstraintGroup((2 * x[i] <= 5 for i in range(2)), name='ubcon')
        self.assertEqual(c1.get_name(), 'ubcon')

    def test_recursive_generation(self):
        G1 = [1, 2]
        G2 = ['a', 'b']
        x = so.VariableGroup(G1, G2, name='x')
        cg = so.ConstraintGroup((x[i, j] >= 1 for i in G1 for j in G2),
                                name='two_indices')
        self.assertEqual(
            repr(cg),
            "sasoptpy.ConstraintGroup([x[1, a] >=  1, x[1, b] >=  1, x[2, a] >=  1, x[2, b] >=  1], name='two_indices')"
        )

        c1 = x[1, 'a'] >= 4
        c2 = x[1, 'b'] <= 6
        cg2 = so.ConstraintGroup([c1, c2], name='list_cons')
        self.assertEqual(
            repr(cg2),
            "sasoptpy.ConstraintGroup([x[1, a] >=  4, x[1, b] <=  6], name='list_cons')"
        )

        def test_type_error():
            cg3 = so.ConstraintGroup({'a': x[1, 'a'] <= 5}, name='errcon')
        self.assertRaises(TypeError, test_type_error)

    def test_expression_mode(self):
        x = so.VariableGroup(2, name='x')
        cg = so.ConstraintGroup((x[i] >= 1 for i in range(2)), name='lbcon')
        exps = cg.get_expressions()
        x_i = exps.loc[0, 'lbcon'].values[0]
        self.assertEqual(x_i.get_value(), 0)

    def test_get_item(self):
        x = so.VariableGroup(2, name='x')
        cg1 = so.ConstraintGroup((x[i] >= 3 for i in range(2)), name='c1')
        self.assertEqual(type(cg1[0]), so.Constraint)

        so.reset()
        I = so.abstract.Set(name='I')
        y = so.VariableGroup(I, name='y')
        cg2 = so.ConstraintGroup((y[i] >= 3 for i in I), name='c2')
        for i in I:
            self.assertEqual(cg2[i]._get_constraint_expr(), "y[o4] >= 3")
        m = so.Model(name='m')
        m.include(cg2)
        m.include(so.abstract.DropStatement(cg2))

    def test_iter(self):
        x = so.VariableGroup(3, name='x')
        cg = so.ConstraintGroup((x[i] >= i for i in range(3)), name='cg')
        for i, con in enumerate(cg):
            self.assertEqual(str(con), "x[{}] >=  {}".format(i, i))

    def test_definition(self):
        x = so.VariableGroup(3, name='x')
        cg = so.ConstraintGroup((x[i] >= 1 for i in range(3)), name='cg')
        self.assertEqual(
            cg._defn(),
            "con cg_0 : x[0] >= 1;\ncon cg_1 : x[1] >= 1;\ncon cg_2 : x[2] >= 1;")

    def test_str(self):
        x = so.VariableGroup(3, name='x')
        cg = so.ConstraintGroup((x[i] <= i-1 for i in range(3)), name='cg')
        cg_str = str(cg)
        self.assertEqual(
            cg_str, cleandoc('''
            Constraint Group (cg) [
              [0: x[0] <=  -1]
              [1: x[1] <=  0]
              [2: x[2] <=  1]
            ]'''))

    def test_invalid_con(self):
        def con_with_no_direction():
            m = so.Model(name='test_cg_invalid_con', session=None)
            x = m.add_variables(3, ub=4, name='x')
            c = m.add_constraints(
                (so.expr_sum((i + j) * x[i] for i in range(3)) for j in
                 range(5)), name='c')

        self.assertRaises(ValueError, con_with_no_direction)

    def test_mixed_case(self):
        m = so.Model(name='m')
        S = so.Set(name='S', value=range(4))
        x = so.VariableGroup(3, name='x')
        y = so.VariableGroup(S, name='y')
        c = so.ConstraintGroup((x[i] + y[j] >= 1 for i in range(3) for j in S), name='c')
        m.include(S, x, y, c)
        assert_equal_wo_temps(self, so.to_optmodel(m), cleandoc('''
            proc optmodel;
               min m_obj = 0;
               set S = 0..3;
               var x {{0,1,2}};
               var y {{S}};
               con c_0 {TEMP1 in S} : x[0] + y[TEMP1] >= 1;
               con c_1 {TEMP2 in S} : x[1] + y[TEMP2] >= 1;
               con c_2 {TEMP3 in S} : x[2] + y[TEMP3] >= 1;
               solve;
            quit;'''))

    def test_members(self):
        x = so.VariableGroup(3, name='x')
        c = so.ConstraintGroup((x[i] <= i for i in range(3)), name='c')
        it = so.Parameter(name='it', value=1)
        e = c[it]
        members = c.get_members()
        self.assertEqual(list(members.keys()), [(0,), (1,), (2,)])
        shadows = c.get_shadow_members()
        self.assertEqual(list(shadows.keys()), [(it,)])
        I = so.Set(name='setI', value=[1])
        i = so.SetIterator(I, name='i')
        f = c[i]
        g = c[i]
        self.assertEqual(id(f), id(g))

    def tearDown(self):
        so.reset()



from .statement_base import Statement
from .literal import LiteralStatement

import sasoptpy


class NestedConditions(Statement):

    def __init__(self):
        super().__init__()

    def append(self, element):
        self.elements.append(element)

    def _defn(self):
        eldefs = []
        for i in self.elements:
            eldefs.append(sasoptpy.to_definition(i))
        return '\n'.join(eldefs)


class IfElseStatement(NestedConditions):

    def __init__(self, logic_expression, if_statement, else_statement=None):
        super().__init__()
        original_container = sasoptpy.container
        c = Case('if', logic_expression)
        sasoptpy.container = c
        if_statement()
        self.append(c)
        if else_statement is not None:
            c = Case('else')
            sasoptpy.container = c
            else_statement()
            self.append(c)
        sasoptpy.container = original_container

    @classmethod
    def if_condition(cls, logic_expression, if_statement, else_statement=None):
        if_else_statement = IfElseStatement(logic_expression, if_statement, else_statement)
        return if_else_statement


class SwitchStatement(NestedConditions):

    def __init__(self, *args):
        super().__init__()
        original_container = sasoptpy.container
        condition = None
        for i in args:
            if sasoptpy.core.util.is_constraint(i):
                condition = i
            elif isinstance(i, sasoptpy.abstract.Condition):
                condition = i
            elif callable(i):
                key = self.get_next_key(condition)
                c = Case(keyword=key, condition=condition)
                sasoptpy.container = c
                i()
                self.append(c)
                condition = None
        sasoptpy.container = original_container

    def get_next_key(self, condition):
        if len(self.elements) == 0:
            return 'if'
        else:
            if sasoptpy.core.util.is_constraint(condition):
                return 'else if'
            elif isinstance(condition, sasoptpy.abstract.Condition):
                return 'else if'
            else:
                return 'else'

    @classmethod
    def switch_condition(cls, *args):
        sc = SwitchStatement(*args)
        return sc


class Case(Statement):

    def __init__(self, keyword='if', condition=None):
        super().__init__()
        self.keyword = keyword
        self.condition = condition

    def get_do_word(self):
        if self.keyword == 'if' or self.keyword == 'else if':
            return 'then do'
        else:
            return 'do'

    def _defn(self):
        s = self.keyword + ' '
        if self.condition:
            s += self.condition._expr() + ' '
        s += self.get_do_word() + ';\n'
        eldefs = []
        for i in self.elements:
            eldef = sasoptpy.to_definition(i)
            if eldef is not None:
                eldefs.append(eldef)
        s += sasoptpy.util.addSpaces('\n'.join(eldefs), 3) + '\n'
        s += 'end;'
        return s

    def append(self, x):
        self.elements.append(x)



# Util functions
from .util import *

# Problem formats
from .format.mps_format import to_mps
from .format.optmodel_format import to_optmodel

# Problem solvers
from .solver.mediator import Mediator
from .solver.cas_mediator import CASMediator
from .solver.sas_mediator import SASMediator



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Model includes :class:`Model` class, the main structure of an opt. model

"""


from collections import OrderedDict
import inspect
from math import inf
from types import GeneratorType
import warnings

import numpy as np
import pandas as pd

import sasoptpy
import sasoptpy.util
from sasoptpy.core import (Expression, Objective, Variable, VariableGroup,
                           Constraint, ConstraintGroup)


class Model:
    """
    Creates an optimization model

    Parameters
    ----------
    name : string
        Name of the model
    session : :class:`swat.cas.connection.CAS` or \
:class:`saspy.SASsession`, optional
        CAS or SAS Session object

    Examples
    --------

    >>> from swat import CAS
    >>> import sasoptpy as so
    >>> s = CAS('cas.server.address', port=12345)
    >>> m = so.Model(name='my_model', session=s)
    NOTE: Initialized model my_model

    >>> mip = so.Model(name='mip')
    NOTE: Initialized model mip
    """

    @sasoptpy.class_containable
    def __init__(self, name=None, session=None):
        self._name = name
        self._objorder = sasoptpy.util.get_creation_id()
        self._session = session

        self._members = {}

        self._variableDict = OrderedDict()
        self._constraintDict = OrderedDict()
        self._objectiveDict = OrderedDict()

        self._setDict = OrderedDict()
        self._parameterDict = OrderedDict()
        self._impvarDict = OrderedDict()
        self._statementDict = OrderedDict()
        self._postSolveDict = OrderedDict()

        self._soltime = 0
        self._objval = None
        self._status = ''
        self._castablename = None
        self._mpsmode = 0
        self._problemSummary = None
        self._solutionSummary = None
        self._primalSolution = None
        self._dualSolution = None
        self._tunerResults = None
        self._milp_opts = {}
        self._lp_opts = {}
        self.response = None

        self._droppedCons = OrderedDict()
        self._droppedVars = OrderedDict()

        self._objective = Objective(0, name=name + '_obj', default=True,
                                    internal=True)

        print('NOTE: Initialized model {}.'.format(name))

    def __eq__(self, other):
        if not isinstance(other, sasoptpy.Model):
            warnings.warn('Cannot compare Model object with {}'.
                          format(type(other)), RuntimeWarning, stacklevel=2)
            return False
        return super().__eq__(other)

    def get_name(self):
        """
        Returns model name
        """
        return self._name

    def add(self, object):
        self.include(object)

    def add_variable(self, name, vartype=None,
                     lb=None, ub=None, init=None):
        """
        Adds a new variable to the model

        New variables can be created via this method or existing variables
        can be added to the model.

        Parameters
        ----------
        name : string
            Name of the variable to be created
        vartype : string, optional
            Type of the variable, either `sasoptpy.BIN`, `sasoptpy.INT` or
            `sasoptpy.CONT`
        lb : float, optional
            Lower bound of the variable
        ub : float, optional
            Upper bound of the variable
        init : float, optional
            Initial value of the variable

        Returns
        -------
        var : :class:`Variable`
            Variable that is added to the model

        Examples
        --------
        Adding a variable on the fly

        >>> m = so.Model(name='demo')
        >>> x = m.add_variable(name='x', vartype=so.INT, ub=10, init=2)
        >>> print(repr(x))
        NOTE: Initialized model demo
        sasoptpy.Variable(name='x', lb=0, ub=10, init=2, vartype='INT')

        Adding an existing variable to a model

        >>> y = so.Variable(name='y', vartype=so.BIN)
        >>> m = so.Model(name='demo')
        >>> m.include(y)

        Notes
        -----

        * `name` is a mandatory field for this method.

        See also
        --------
        :class:`Variable`, :func:`Model.include`
        """

        var = Variable(name, vartype, lb, ub, init)
        self.include(var)
        return var

    def add_variables(self, *argv, name,
                      vartype=None,
                      lb=None, ub=None, init=None):
        """
        Adds a group of variables to the model

        Parameters
        ----------
        argv : list, dict, :class:`pandas.Index`
            Loop index for variable group
        name : string
            Name of the variables
        vartype : string, optional
            Type of variables, `BIN`, `INT`, or `CONT`
        lb : list, dict, :class:`pandas.Series`
            Lower bounds of variables
        ub : list, dict, :class:`pandas.Series`
            Upper bounds of variables
        init : list, dict, :class:`pandas.Series`
            Initial values of variables

        See also
        --------
        :class:`VariableGroup`, :meth:`Model.include`

        Examples
        --------

        >>> production = m.add_variables(PERIODS, vartype=so.INT,
                                        name='production', lb=min_production)
        >>> print(production)
        >>> print(repr(production))
        Variable Group (production) [
          [Period1: production['Period1',]]
          [Period2: production['Period2',]]
          [Period3: production['Period3',]]
        ]
        sasoptpy.VariableGroup(['Period1', 'Period2', 'Period3'],
        name='production')

        """

        vg = VariableGroup(*argv, name=name, vartype=vartype, lb=lb, ub=ub,
                           init=init)
        self.include(vg)
        return vg

    @sasoptpy.containable
    def add_constraint(self, c, name):
        """
        Adds a single constraint to the model

        Parameters
        ----------
        c : :class:`Constraint`
            Constraint to be added to the model
        name : string
            Name of the constraint

        Returns
        -------
        c : :class:`Constraint`
            Reference to the constraint

        Examples
        --------

        >>> x = m.add_variable(name='x', vartype=so.INT, lb=0, ub=5)
        >>> y = m.add_variables(3, name='y', vartype=so.CONT, lb=0, ub=10)
        >>> c1 = m.add_constraint(x + y[0] >= 3, name='c1')
        >>> print(c1)
         x  +  y[0]  >=  3

        >>> c2 = m.add_constraint(x - y[2] == [4, 10], name='c2')
        >>> print(c2)
         -  y[2]  +  x  =  [4, 10]

        See also
        --------
        :class:`Constraint`, :meth:`Model.include`

        """

        if ((c._direction == 'L' and c._linCoef['CONST']['val'] == -inf) or
           (c._direction == 'G' and c._linCoef['CONST']['val'] == inf)):
            raise ValueError("Invalid constant value for the constraint type")

        if c._name is None:
            c.set_name(name)
            c.set_permanent()

        self.include(c)
        return c

    def add_constraints(self, argv, name):
        """
        Adds a set of constraints to the model

        Parameters
        ----------
        argv : Generator-type object
            List of constraints as a generator-type Python object
        name : string
            Name for the constraint group and individual constraint prefix

        Returns
        -------
        cg : :class:`ConstraintGroup`
            Reference to the ConstraintGroup

        Examples
        --------

        >>> x = m.add_variable(name='x', vartype=so.INT, lb=0, ub=5)
        >>> y = m.add_variables(3, name='y', vartype=so.CONT, lb=0, ub=10)
        >>> c = m.add_constraints((x + 2 * y[i] >= 2 for i in [0, 1, 2]),
                                  name='c')
        >>> print(c)
        Constraint Group (c) [
          [0:  2.0 * y[0]  +  x  >=  2]
          [1:  2.0 * y[1]  +  x  >=  2]
          [2:  2.0 * y[2]  +  x  >=  2]
        ]

        >>> t = m.add_variables(3, 4, name='t')
        >>> ct = m.add_constraints((t[i, j] <= x for i in range(3)
                                   for j in range(4)), name='ct')
        >>> print(ct)
        Constraint Group (ct) [
          [(0, 0):  -  x  +  t[0, 0]  <=  0]
          [(0, 1):  t[0, 1]  -  x  <=  0]
          [(0, 2):  -  x  +  t[0, 2]  <=  0]
          [(0, 3):  t[0, 3]  -  x  <=  0]
          [(1, 0):  t[1, 0]  -  x  <=  0]
          [(1, 1):  t[1, 1]  -  x  <=  0]
          [(1, 2):  -  x  +  t[1, 2]  <=  0]
          [(1, 3):  -  x  +  t[1, 3]  <=  0]
          [(2, 0):  -  x  +  t[2, 0]  <=  0]
          [(2, 1):  t[2, 1]  -  x  <=  0]
          [(2, 2):  t[2, 2]  -  x  <=  0]
          [(2, 3):  t[2, 3]  -  x  <=  0]
        ]

        See also
        --------
        :class:`ConstraintGroup`, :meth:`Model.include`

        """

        if type(argv) == list or type(argv) == GeneratorType:
            cg = ConstraintGroup(argv, name=name)
            self.include(cg)
            return cg
        elif sasoptpy.core.util.is_constraint(argv):
            warnings.warn(
                'Use add_constraint method for adding single constraints',
                UserWarning)
            c = self.add_constraint(argv, name=name)
            return c

    def add_set(self, name, init=None, value=None, settype=None):
        """
        Adds a set to the model

        Parameters
        ----------
        name : string, optional
            Name of the set
        init : :class:`Set`, optional
            Initial value of the set
        value : list, float, optional
            Exact value of the set
        settype : list, optional
            Types of the set as a list

            The list can have one more `num` (for float) and `str` (for string)
            values. You can use `sasoptpy.NUM` and `sasoptpy.STR` for floats and
            strings, respectively.

        Examples
        --------

        >>> I = m.add_set(name='I')
        >>> print(I._defn())
        set I;

        >>> J = m.add_set(name='J', settype=['str'])
        >>> print(J._defn())
        set <str> J;

        >>> N = m.add_parameter(name='N', init=4)
        >>> K = m.add_set(name='K', init=so.exp_range(1, N))
        >>> print(K._defn())
        set K = 1..N;

        >>> m.add_set(name='W', settype=[so.STR, so.NUM])
        >>> print(W._defn())
        set <str, num> W;

        """
        new_set = sasoptpy.abstract.Set(name, init=init, value=value,
                                        settype=settype)
        self.include(new_set)
        return new_set

    def add_parameter(self, *argv, name, init=None, value=None, p_type=None):
        """
        Adds a :class:`abstract.Parameter` object to the model

        Parameters
        ----------
        argv : :class:`Set`, optional
            Index or indices of the parameter
        name : string
            Name of the parameter
        init : float or expression, optional
            Initial value of the parameter
        p_type : string, optional
            Type of the parameter, 'num' for floats or 'str' for strings

        Examples
        --------

        >>> I = m.add_set(name='I')
        >>> a = m.add_parameter(I, name='a', init=5)
        >>> print(a._defn())
        num a {I} init 5 ;

        >>> I = m.add_set(name='I')
        >>> J = m.add_set(name='J')
        >>> p = m.add_parameter(I, J, name='p')
        >>> print(p._defn())
        num p {{I,J}};

        Returns
        -------

        p : :class:`abstract.Parameter` or :class:`abstract.ParameterGroup`
            A single parameter or a parameter group

        """
        if len(argv) == 0:
            p = sasoptpy.abstract.Parameter(
                name, init=init, value=value, ptype=p_type)
            self.include(p)
            return p
        else:
            keylist = list(argv)
            p = sasoptpy.abstract.ParameterGroup(keylist, name=name, init=init,
                                                 value=value, ptype=p_type)
            self.include(p)
            return p

    def add_implicit_variable(self, argv=None, name=None):
        """
        Adds an implicit variable to the model

        Parameters
        ----------
        argv : Generator-type object
            Generator object where each item is an entry
        name : string
            Name of the implicit variable

        Examples
        --------

        >>> x = m.add_variables(range(5), name='x')
        >>> y = m.add_implicit_variable((
        >>>     x[i] + 2 * x[i+1] for i in range(4)), name='y')
        >>> print(y[2])
        x[2] + 2 * x[3]

        >>> I = m.add_set(name='I')
        >>> z = m.add_implicit_variable((x[i] * 2 + 2 for i in I), name='z')
        >>> print(z._defn())
        impvar z {i_1 in I} = 2 * x[i_1] + 2;


        Notes
        -----

        - Based on whether the implicit variables  are generated by a regular
          or abstract expression, they can appear in generated OPTMODEL codes.

        """
        if name is None:
            name = sasoptpy.util.get_next_name()
        iv = sasoptpy.abstract.ImplicitVar(argv=argv, name=name)
        self.include(iv)
        return iv

    def add_statement(self, statement, after_solve=None):
        """
        Adds a PROC OPTMODEL statement to the model

        Parameters
        ----------
        statement : :class:`Expression` or string
            Statement object
        after_solve : boolean
            Switch for appending the statement after the problem solution

        Examples
        --------

        >>> I = m.add_set(name='I')
        >>> x = m.add_variables(I, name='x', vartype=so.INT)
        >>> a = m.add_parameter(I, name='a')
        >>> c = m.add_constraints((x[i] <= 2 * a[i] for i in I), name='c')
        >>> m.add_statement('print x;', after_solve=True)
        >>> print(m.to_optmodel())
        proc optmodel;
        min m_obj = 0;
        set I;
        var x {I} integer >= 0;
        num a {I};
        con c {i_1 in I} : x[i_1] - 2.0 * a[i_1] <= 0;
        solve;
        print _var_.name _var_.lb _var_.ub _var_ _var_.rc;
        print _con_.name _con_.body _con_.dual;
        print x;
        quit;


        Notes
        -----

        - If the statement string includes 'print', then the statement is
          automatically placed after the solve even if `after_solve` is `False`.

        """
        if isinstance(statement, sasoptpy.abstract.Statement):
            self._save_statement(statement, after_solve)
        elif isinstance(statement, str):
            s = sasoptpy.abstract.LiteralStatement(statement)
            self._save_statement(s, after_solve)

    def add_postsolve_statement(self, statement):
        if isinstance(statement, sasoptpy.abstract.Statement):
            self._save_statement(statement, after_solve=True)
        elif isinstance(statement, str):
            s = sasoptpy.abstract.LiteralStatement(statement)
            self._save_statement(s, after_solve=True)

    def _save_statement(self, st, after_solve=None):
        if after_solve is None or after_solve is False:
            self._statementDict[id(st)] = st
        else:
            self._postSolveDict[id(st)] = st

    def drop_variable(self, variable):
        """
        Drops a variable from the model

        Parameters
        ----------
        variable : :class:`Variable`
            The variable to be dropped from the model

        Examples
        --------

        >>> x = m.add_variable(name='x')
        >>> y = m.add_variable(name='y')
        >>> print(m.get_variable('x'))
         x
        >>> m.drop_variable(x)
        >>> print(m.get_variable('x'))
        None

        See also
        --------
        :func:`Model.drop_variables`
        :func:`Model.drop_constraint`
        :func:`Model.drop_constraints`

        """
        vname = variable.get_name()
        if self._variableDict.pop(vname, None) is None:
            self._droppedVars[vname] = True

    def restore_variable(self, variable):
        vname = variable.get_name()
        if variable.get_parent_reference()[0] is not None:
            self._droppedVars.pop(vname, None)
        else:
            self.include(variable)

    @sasoptpy.containable
    def drop_constraint(self, constraint):
        """
        Drops a constraint from the model

        Parameters
        ----------
        constraint : :class:`Constraint`
            The constraint to be dropped from the model

        Examples
        --------

        >>> c1 = m.add_constraint(2 * x + y <= 15, name='c1')
        >>> print(m.get_constraint('c1'))
        2 * x  +  y  <=  15
        >>> m.drop_constraint(c1)
        >>> print(m.get_constraint('c1'))
        None

        See also
        --------
        :func:`Model.drop_constraints`
        :func:`Model.drop_variable`
        :func:`Model.drop_variables`

        """

        cname = constraint.get_name()
        if self._constraintDict.pop(cname, None) is None:
            self._droppedCons[constraint._get_optmodel_name()] = True

    def restore_constraint(self, constraint):
        cname = constraint.get_name()
        if constraint.get_parent_reference()[0] is not None:
            self._droppedCons.pop(constraint._get_optmodel_name(), None)
        else:
            self.include(constraint)

    def drop_variables(self, *variables):
        """
        Drops a variable group from the model

        Parameters
        ----------
        variables : :class:`VariableGroup`
            The variable group to be dropped from the model

        Examples
        --------

        >>> x = m.add_variables(3, name='x')
        >>> print(m.get_variables())
        [sasoptpy.Variable(name='x_0',  vartype='CONT'),
         sasoptpy.Variable(name='x_1',  vartype='CONT')]
        >>> m.drop_variables(x)
        >>> print(m.get_variables())
        []

        See also
        --------
        :func:`Model.drop_variable`
        :func:`Model.drop_constraint`
        :func:`Model.drop_constraints`

        """
        for v in variables:
            self.drop_variable(v)

    def drop_constraints(self, *constraints):
        """
        Drops a constraint group from the model

        Parameters
        ----------
        constraints : :class:`Constraint` or :class:`ConstraintGroup`
            Arbitrary number of constraints to be dropped

        Examples
        --------

        >>> c1 = m.add_constraints((x[i] + y <= 15 for i in [0, 1]), name='c1')
        >>> print(m.get_constraints())
        [sasoptpy.Constraint( x[0]  +  y  <=  15, name='c1_0'),
         sasoptpy.Constraint( x[1]  +  y  <=  15, name='c1_1')]
        >>> m.drop_constraints(c1)
        >>> print(m.get_constraints())
        []

        See also
        --------
        :func:`Model.drop_constraints`
        :func:`Model.drop_variable`
        :func:`Model.drop_variables`

        """
        for c in constraints:
            self.drop_constraint(c)

    def include(self, *argv):
        """
        Adds existing variables and constraints to a model

        Parameters
        ----------
        argv :
            Objects to be included in the model

        Notes
        -----
        * Valid object types for `argv` parameter:

          - :class:`Model`

            Including a model causes all variables and constraints inside the
            original model to be included.

          - :class:`Variable`
          - :class:`Constraint`
          - :class:`VariableGroup`
          - :class:`ConstraintGroup`
          - :class:`Objective`
          - :class:`Set`
          - :class:`Parameter`
          - :class:`ParameterGroup`
          - :class:`Statement` and all subclasses
          - :class:`ImplicitVar`


        Examples
        --------

        Adding an existing variable

        >>> x = so.Variable(name='x', vartype=so.CONT)
        >>> m.include(x)

        Adding an existing constraint

        >>> c1 = so.Constraint(x + y <= 5, name='c1')
        >>> m.include(c1)

        Adding an existing set of variables

        >>> z = so.VariableGroup(3, 5, name='z', ub=10)
        >>> m.include(z)

        Adding an existing set of constraints

        >>> c2 = so.ConstraintGroup((x + 2 * z[i, j] >= 2 for i in range(3)
                                    for j in range(5)), name='c2')
        >>> m.include(c2)

        Adding an existing model (including all of its elements)

        >>> new_model = so.Model(name='new_model')
        >>> new_model.include(m)

        """

        include_methods = {
            Variable: self._include_variable,
            VariableGroup: self._include_vargroup,
            Constraint: self._include_constraint,
            ConstraintGroup: self._include_congroup,
            Objective: self._set_objective,
            sasoptpy.Set: self._include_set,
            sasoptpy.Parameter: self._include_parameter,
            sasoptpy.ParameterGroup: self._include_parameter_group,
            sasoptpy.abstract.LiteralStatement: self._include_statement,
            sasoptpy.ImplicitVar: self._include_expdict,
            sasoptpy.abstract.ReadDataStatement: self._include_statement,
            sasoptpy.abstract.DropStatement: self._include_statement,
            list: self.include,
            Model: self._include_model
        }

        for c in argv:
            meth = include_methods.get(type(c))
            if any(isinstance(c, i) for i in [Variable, VariableGroup, Constraint, ConstraintGroup, Objective]):
                if sasoptpy.container is not None:
                    if c._objorder > self._objorder:
                        raise ReferenceError('Object {} should be defined before Model {} inside a Workspace'.format(
                            c._expr(), self.get_name()
                    ))
            if meth is not None:
                meth(c)

    def _include_variable(self, var):
        vname = var.get_name()
        if vname in self._variableDict:
            warnings.warn(f"Variable name {vname} exists in the model."
                          "New declaration will override the existing value.",
                          UserWarning)
        self._variableDict[vname] = var

    def _include_vargroup(self, vg):
        self._variableDict[vg.get_name()] = vg

    def _include_constraint(self, con):
        if sasoptpy.core.util.has_parent(con):
            return
        name = con.get_name()
        if con.get_name() in self._constraintDict:
            warnings.warn(f"Constraint name {name} exists in the model."
                          "New declaration will override the existing value.",
                          UserWarning)
        self._constraintDict[con.get_name()] = con

    def _include_congroup(self, cg):
        self._constraintDict[cg.get_name()] = cg

    def _set_objective(self, ob):
        self._objective = ob

    def _include_set(self, st):
        self._setDict[id(st)] = st

    def _include_parameter(self, p):
        self._parameterDict[p.get_name()] = p

    def _include_parameter_group(self, pg):
        self._parameterDict[pg.get_name()] = pg

    def _include_statement(self, os):
        self._save_statement(os)

    def _include_expdict(self, ed):
        self._impvarDict[ed.get_name()] = ed

    def _include_model(self, model):
        self._setDict.update(model._setDict)
        self._parameterDict.update(model._parameterDict)
        self._statementDict.update(model._statementDict)
        self._postSolveDict.update(model._postSolveDict)
        self._impvarDict.update(model._impvarDict)
        for s in model.get_grouped_variables().values():
            self._include_variable(s)
        for s in model.get_grouped_constraints().values():
            self._include_constraint(s)
        self._objective = model._objective

    def drop(self, obj):
        if isinstance(obj, sasoptpy.VariableGroup):
            self.drop_variables(obj)
        elif isinstance(obj, sasoptpy.Variable):
            self.drop_variable(obj)
        elif isinstance(obj, sasoptpy.ConstraintGroup):
            self.drop_constraints(obj)
        elif isinstance(obj, sasoptpy.Constraint):
            self.drop_constraint(obj)
        elif isinstance(obj, sasoptpy.Set):
            self._setDict.pop(id(obj), None)
        elif isinstance(obj, sasoptpy.Parameter) or\
             isinstance(obj, sasoptpy.ParameterGroup):
            self._parameterDict.pop(obj.get_name(), None)
        elif isinstance(obj, sasoptpy.abstract.Statement):
            self._statementDict.pop(id(obj), None)

    def set_objective(self, expression, name, sense=None):
        """
        Specifies the objective function for the model

        Parameters
        ----------
        expression : :class:`Expression`
            The objective function as an Expression
        name : string
            Name of the objective value
        sense : string, optional
            Objective value direction, `sasoptpy.MIN` or `sasoptpy.MAX`

        Returns
        -------
        objective : :class:`Expression`
            Objective function as an :class:`Expression` object

        Examples
        --------

        >>> profit = so.Expression(5 * sales - 2 * material, name='profit')
        >>> m.set_objective(profit, so.MAX)
        >>> print(m.get_objective())
         -  2.0 * material  +  5.0 * sales

        >>> m.set_objective(4 * x - 5 * y, name='obj')
        >>> print(repr(m.get_objective()))
        sasoptpy.Expression(exp =  4.0 * x  -  5.0 * y , name='obj')

        >>> f1 = m.set_objective(2 * x + y, sense=so.MIN, name='f1')
        >>> f2 = m.append_objective( (x - y) ** 2, sense=so.MIN, name='f2')
        >>> print(m.to_optmodel(options={'with': 'blackbox', 'obj': (f1, f2)}))
        proc optmodel;
        var x;
        var y;
        min f1 = 2 * x + y;
        min f2 = (x - y) ^ (2);
        solve with blackbox obj (f1 f2);
        print _var_.name _var_.lb _var_.ub _var_ _var_.rc;
        print _con_.name _con_.body _con_.dual;
        quit;

        Notes
        -----
        - Default objective sense is minimization `MIN`.
        - This method replaces the existing objective of the model.
          When working with multiple objectives, use the
          :meth:`Model.append_objective` method.

        See also
        --------
        :meth:`Model.append_objective`

        """

        obj = Objective(expression, sense=sense, name=name)
        self._objective = obj
        return self._objective

    def append_objective(self, expression, name, sense=None):
        """
        Appends a new objective to the model

        Parameters
        ----------
        expression : :class:`Expression`
            The objective function as an Expression
        name : string
            Name of the objective value
        sense : string, optional
            Objective value direction, `sasoptpy.MIN` or `sasoptpy.MAX`

        Returns
        -------
        objective : :class:`Expression`
            Objective function as an :class:`Expression` object

        Examples
        --------

        >>> f1 = m.set_objective(2 * x + y, sense=so.MIN, name='f1')
        >>> f2 = m.append_objective( (x - y) ** 2, sense=so.MIN, name='f2')
        >>> print(m.to_optmodel(options={'with': 'blackbox', 'obj': (f1, f2)}))
        proc optmodel;
        var x;
        var y;
        min f1 = 2 * x + y;
        min f2 = (x - y) ^ (2);
        solve with blackbox obj (f1 f2);
        print _var_.name _var_.lb _var_.ub _var_ _var_.rc;
        print _con_.name _con_.body _con_.dual;
        quit;

        Notes
        -----
        - Default objective sense is minimization `MIN`.

        See also
        --------
        :meth:`Model.set_objective`

        """
        obj = Objective(expression, name=name, sense=sense)
        self._objectiveDict[id(obj)] = obj
        return obj

    def get_objective(self):
        """
        Returns the objective function as an :class:`Expression` object

        Returns
        -------
        objective : :class:`Expression`
            Objective function

        Examples
        --------

        >>> m.set_objective(4 * x - 5 * y, name='obj')
        >>> print(repr(m.get_objective()))
        sasoptpy.Expression(exp =  4.0 * x  -  5.0 * y , name='obj')

        """
        return self._objective

    def get_all_objectives(self):
        """
        Returns a list of objectives in the model

        Returns
        -------
        all_objectives : list
           A list of :class:`Objective` objects

        Examples
        --------

        >>> m = so.Model(name='test_set_get_objective')
        >>> x = m.add_variable(name='x')
        >>> obj1 = m.set_objective(2 * x, sense=so.MIN, name='obj1')
        >>> obj2 = m.set_objective(5 * x, sense=so.MIN, name='obj2') # Overrides obj1
        >>> obj3 = m.append_objective(10 * x, sense=so.MIN, name='obj3')
        >>> assertEqual(m.get_all_objectives(), [obj2, obj3])
        True

        """
        all_objs = list(self._objectiveDict.values())
        all_objs.append(self._objective)
        return sorted(all_objs, key=lambda i: i._objorder)

    def get_objective_value(self):
        """
        Returns the optimal objective value

        Returns
        -------
        objective_value : float
            Optimal objective value at current solution

        Examples
        --------

        >>> m.solve()
        >>> print(m.get_objective_value())
        42.0

        Notes
        -----

        - This method should be used for getting the objective value after
          solve.
        - In order to get the current value of the objective after changing
          variable values, you can use :code:`m.get_objective().get_value()`.

        """
        if self._objval is not None:
            return sasoptpy.util.get_in_digit_format(self._objval)
        else:
            return self.get_objective().get_value()

    def set_objective_value(self, value):
        self._objval = value

    def get_constraint(self, name):
        """
        Returns the reference to a constraint in the model

        Parameters
        ----------
        name : string
            Name of the constraint requested

        Returns
        -------
        constraint : :class:`Constraint`
            Requested object

        Examples
        --------

        >>> m.add_constraint(2 * x + y <= 15, name='c1')
        >>> print(m.get_constraint('c1'))
        2.0 * x  +  y  <=  15

        """
        # return self._constraintDict.get(name, None)
        constraints = self.get_constraints_dict()
        safe_name = name.replace('\'', '')
        if name in constraints:
            return constraints[name]
        elif safe_name in constraints:
            return constraints[safe_name]
        elif '[' in name:
            first_part = name.split('[')[0]
            if constraints.get(first_part, None) is not None:
                vg = constraints.get(first_part)
                return vg.get_member_by_name(name)
        else:
            return None

    def loop_constraints(self):
        for i in self._constraintDict.values():
            if isinstance(i, sasoptpy.Constraint):
                yield i
            elif isinstance(i, sasoptpy.ConstraintGroup):
                for j in i.get_members().values():
                    yield j

    def _get_all_constraints(self):
        all_cons = OrderedDict()
        for c in self._constraintDict.values():
            if isinstance(c, sasoptpy.Constraint):
                all_cons[c.get_name()] = c
            elif isinstance(c, sasoptpy.ConstraintGroup):
                for sc in c.get_members().values():
                    all_cons[sc.get_name()] = sc
        return all_cons

    def get_constraints(self):
        """
        Returns a list of constraints in the model

        Returns
        -------
        constraints : list
            A list of Constraint objects

        Examples
        --------

        >>> m.add_constraint(x[0] + y <= 15, name='c1')
        >>> m.add_constraints((2 * x[i] - y >= 1 for i in [0, 1]), name='c2')
        >>> print(m.get_constraints())
        [sasoptpy.Constraint( x[0]  +  y  <=  15, name='c1'),
         sasoptpy.Constraint( 2.0 * x[0]  -  y  >=  1, name='c2_0'),
         sasoptpy.Constraint( 2.0 * x[1]  -  y  >=  1, name='c2_1')]

        """
        return list(self.loop_constraints())

    def get_constraints_dict(self):
        return self._constraintDict

    def get_grouped_constraints(self):
        """
        Returns an ordered dictionary of constraints

        Returns
        -------
        grouped_cons : OrderedDict
           Dictionary of constraints and constraint groups in the model

        Examples
        --------

        >>> m1 = so.Model(name='test_copy_model_1')
        >>> x = m1.add_variable(name='x')
        >>> y = m1.add_variables(2, name='y')
        >>> c1 = m1.add_constraint(x + y[0] >= 2, name='c1')
        >>> c2 = m1.add_constraints((x - y[i] <= 10 for i in range(2)), name='c2')
        >>> cons = OrderedDict([('c1', c1), ('c2', c2)])
        >>> self.assertEqual(m1.get_grouped_constraints(), cons)
        True

        See also
        --------
        :meth:`Model.get_constraints`, :meth:`Model.get_grouped_variables`

        """
        return self.get_constraints_dict()

    def get_variable(self, name):
        """
        Returns the reference to a variable in the model

        Parameters
        ----------
        name : string
            Name or key of the variable requested

        Returns
        -------
        variable : :class:`Variable`
            Reference to the variable

        Examples
        --------

        >>> m.add_variable(name='x', vartype=so.INT, lb=3, ub=5)
        >>> var1 = m.get_variable('x')
        >>> print(repr(var1))
        sasoptpy.Variable(name='x', lb=3, ub=5, vartype='INT')

        """
        variables = self.get_variable_dict()
        safe_name = name.replace('\'', '')
        if name in variables:
            return variables[name]
        elif safe_name in variables:
            return variables[safe_name]
        elif '[' in name:
            first_part = name.split('[')[0]
            if variables.get(first_part, None) is not None:
                vg = variables.get(first_part)
                return vg.get_member_by_name(name)
        else:
            return None

    def loop_variables(self):
        for i in self._variableDict.values():
            if isinstance(i, sasoptpy.Variable):
                yield i
            elif isinstance(i, sasoptpy.VariableGroup):
                for j in i.get_members().values():
                    yield j

    def _get_all_variables(self):
        all_vars = OrderedDict()
        for v in self._variableDict.values():
            if isinstance(v, sasoptpy.Variable):
                all_vars[v.get_name()] = v
            elif isinstance(v, sasoptpy.VariableGroup):
                for sc in v.get_members().values():
                    all_vars[sc.get_name()] = sc
        return all_vars

    def get_variables(self):
        """
        Returns a list of variables

        Returns
        -------
        variables : list
            List of variables in the model

        Examples
        --------

        >>> x = m.add_variables(2, name='x')
        >>> y = m.add_variable(name='y')
        >>> print(m.get_variables())
        [sasoptpy.Variable(name='x_0',  vartype='CONT'),
         sasoptpy.Variable(name='x_1',  vartype='CONT'),
         sasoptpy.Variable(name='y',  vartype='CONT')]

        """
        return list(self.loop_variables())

    def get_variable_dict(self):
        return self._variableDict

    def get_grouped_variables(self):
        """
        Returns an ordered dictionary of variables

        Returns
        -------
        grouped_vars : OrderedDict
           Dictionary of variables and variable groups in the model

        Examples
        --------

        >>> m1 = so.Model(name='test_copy_model_1')
        >>> x = m1.add_variable(name='x')
        >>> y = m1.add_variables(2, name='y')
        >>> vars = OrderedDict([('x', x), ('y', y)])
        >>> self.assertEqual(m1.get_grouped_variables(), vars)
        True

        See also
        --------
        :meth:`Model.get_variables`, :meth:`Model.get_grouped_constraints`

        """
        return self.get_variable_dict()

    def get_variable_coef(self, var):
        """
        Returns the objective value coefficient of a variable

        Parameters
        ----------
        var : :class:`Variable` or string
            Variable whose objective value is requested. It can be either the
            variable object itself, or the name of the variable.

        Returns
        -------
        coef : float
            Objective value coefficient of the given variable

        Examples
        --------

        >>> x = m.add_variable(name='x')
        >>> y = m.add_variable(name='y')
        >>> m.set_objective(4 * x - 5 * y, name='obj', sense=so.MAX)
        >>> print(m.get_variable_coef(x))
        4.0
        >>> print(m.get_variable_coef('y'))
        -5.0

        """
        if isinstance(var, sasoptpy.core.Variable):
            varname = var.get_name()
        else:
            varname = var
        if varname in self._objective._linCoef:
            return self._objective._linCoef[varname]['val']
        else:
            if self.get_objective()._is_linear():
                if varname in self._variableDict:
                    return 0
                else:
                    raise RuntimeError('Variable is not a member of the model')
            else:
                warnings.warn('Objective is not linear', RuntimeWarning)

    def set_variable_coef(self, var, coef):
        varname = var.get_name()
        if varname in self._objective._linCoef:
            self._objective._linCoef[varname]['val'] = coef
        else:
            self._objective += coef*var

    def set_variable_value(self, name, value):

        variable = self.get_variable(name)
        if variable is not None:
            variable.set_value(value)
        else:
            self._set_abstract_values(name, value)

    def set_dual_value(self, name, value):
        variable = self.get_variable(name)
        if variable is not None:
            variable.set_dual(value)

    def get_variable_value(self, var):
        """
        Returns the value of a variable

        Parameters
        ----------
        var : :class:`Variable` or string
            Variable reference

        Notes
        -----
        - It is possible to get a variable's value by using the
          :func:`Variable.get_value` method, as long as the variable is not
          abstract.
        - This method is a wrapper around :func:`Variable.get_value` and an
          overlook function for model components.
        """

        if sasoptpy.core.util.is_variable(var):
            varname = var.get_name()
        else:
            varname = var

        if varname in self._variableDict:
            return self._variableDict[varname].get_value()
        else:
            return self._get_variable_solution(varname)

    def _get_variable_solution(self, name):
        if self._primalSolution is not None:
            for row in self._primalSolution.itertuples():
                if row.var == name:
                    return row.value
        else:
            raise RuntimeError('No primal solution is available')

        warnings.warn('Variable could not be found')
        return None

    def get_problem_summary(self):
        """
        Returns the problem summary table to the user

        Returns
        -------
        ps : :class:`swat.dataframe.SASDataFrame`
            Problem summary table, that is obtained after :meth:`Model.solve`

        Examples
        --------

        >>> m.solve()
        >>> ps = m.get_problem_summary()
        >>> print(type(ps))
        <class 'swat.dataframe.SASDataFrame'>

        >>> print(ps)
        Problem Summary
                                        Value
        Label
        Problem Name                   model1
        Objective Sense          Maximization
        Objective Function                obj
        RHS                               RHS
        Number of Variables                 2
        Bounded Above                       0
        Bounded Below                       2
        Bounded Above and Below             0
        Free                                0
        Fixed                               0
        Number of Constraints               2
        LE (<=)                             1
        EQ (=)                              0
        GE (>=)                             1
        Range                               0
        Constraint Coefficients             4

        >>> print(ps.index)
        Index(['Problem Name', 'Objective Sense', 'Objective Function', 'RHS',
        '', 'Number of Variables', 'Bounded Above', 'Bounded Below',
        'Bounded Above and Below', 'Free', 'Fixed', '',
        'Number of Constraints', 'LE (<=)', 'EQ (=)', 'GE (>=)', 'Range', '',
        'Constraint Coefficients'],
        dtype='object', name='Label')

        >>> print(ps.loc['Number of Variables'])
        Value               2
        Name: Number of Variables, dtype: object

        >>> print(ps.loc['Constraint Coefficients', 'Value'])
        4

        """
        return self._problemSummary

    def get_solution_summary(self):
        """
        Returns the solution summary table to the user

        Returns
        -------
        ss : :class:`swat.dataframe.SASDataFrame`
            Solution summary table, that is obtained after :meth:`Model.solve`

        Examples
        --------

        >>> m.solve()
        >>> soln = m.get_solution_summary()
        >>> print(type(soln))
        <class 'swat.dataframe.SASDataFrame'>

        >>> print(soln)
        Solution Summary
                                       Value
        Label
        Solver                            LP
        Algorithm               Dual Simplex
        Objective Function               obj
        Solution Status              Optimal
        Objective Value                   10
        Primal Infeasibility               0
        Dual Infeasibility                 0
        Bound Infeasibility                0
        Iterations                         2
        Presolve Time                   0.00
        Solution Time                   0.01

        >>> print(soln.index)
        Index(['Solver', 'Algorithm', 'Objective Function', 'Solution Status',
               'Objective Value', '', 'Primal Infeasibility',
               'Dual Infeasibility', 'Bound Infeasibility', '', 'Iterations',
               'Presolve Time', 'Solution Time'],
              dtype='object', name='Label')

        >>> print(soln.loc['Solution Status', 'Value'])
        Optimal

        """
        return self._solutionSummary

    def get_solution(self, vtype='Primal', solution=None, pivot=False):
        """
        Returns the primal and dual problem solutions

        Parameters
        ----------
        vtype : string, optional
            `Primal` or `Dual`
        solution : integer, optional
            Solution number to be returned (for the MILP solver)
        pivot : boolean, optional
            When set to `True`, returns multiple solutions in columns as a pivot
            table

        Returns
        -------
        solution : :class:`pandas.DataFrame`
            Primal or dual solution table returned from the CAS action

        Examples
        --------

        >>> m.solve()
        >>> print(m.get_solution('Primal'))
                     var   lb             ub  value  solution
        0       x[clock]  0.0  1.797693e+308    0.0       1.0
        1          x[pc]  0.0  1.797693e+308    5.0       1.0
        2   x[headphone]  0.0  1.797693e+308    2.0       1.0
        3         x[mug]  0.0  1.797693e+308    0.0       1.0
        4        x[book]  0.0  1.797693e+308    0.0       1.0
        5         x[pen]  0.0  1.797693e+308    1.0       1.0
        6       x[clock]  0.0  1.797693e+308    0.0       2.0
        7          x[pc]  0.0  1.797693e+308    5.0       2.0
        8   x[headphone]  0.0  1.797693e+308    2.0       2.0
        9         x[mug]  0.0  1.797693e+308    0.0       2.0
        10       x[book]  0.0  1.797693e+308    0.0       2.0
        11        x[pen]  0.0  1.797693e+308    0.0       2.0
        12      x[clock]  0.0  1.797693e+308    1.0       3.0
        13         x[pc]  0.0  1.797693e+308    4.0       3.0
        ...

        >>> print(m.get_solution('Primal', solution=2))
                     var   lb             ub  value  solution
        6       x[clock]  0.0  1.797693e+308    0.0       2.0
        7          x[pc]  0.0  1.797693e+308    5.0       2.0
        8   x[headphone]  0.0  1.797693e+308    2.0       2.0
        9         x[mug]  0.0  1.797693e+308    0.0       2.0
        10       x[book]  0.0  1.797693e+308    0.0       2.0
        11        x[pen]  0.0  1.797693e+308    0.0       2.0

        >>> print(m.get_solution(pivot=True))
        solution      1.0  2.0  3.0  4.0  5.0
        var
        x[book]       0.0  0.0  0.0  1.0  0.0
        x[clock]      0.0  0.0  1.0  1.0  0.0
        x[headphone]  2.0  2.0  1.0  1.0  0.0
        x[mug]        0.0  0.0  0.0  1.0  0.0
        x[pc]         5.0  5.0  4.0  1.0  0.0
        x[pen]        1.0  0.0  0.0  1.0  0.0

        >>> print(m.get_solution('Dual'))
                             con  value  solution
        0             weight_con   20.0       1.0
        1       limit_con[clock]    0.0       1.0
        2          limit_con[pc]    5.0       1.0
        3   limit_con[headphone]    2.0       1.0
        4         limit_con[mug]    0.0       1.0
        5        limit_con[book]    0.0       1.0
        6         limit_con[pen]    1.0       1.0
        7             weight_con   19.0       2.0
        8       limit_con[clock]    0.0       2.0
        9          limit_con[pc]    5.0       2.0
        10  limit_con[headphone]    2.0       2.0
        11        limit_con[mug]    0.0       2.0
        12       limit_con[book]    0.0       2.0
        13        limit_con[pen]    0.0       2.0
        ...

        >>> print(m.get_solution('dual', pivot=True))
        solution               1.0   2.0   3.0   4.0  5.0
        con
        limit_con[book]        0.0   0.0   0.0   1.0  0.0
        limit_con[clock]       0.0   0.0   1.0   1.0  0.0
        limit_con[headphone]   2.0   2.0   1.0   1.0  0.0
        limit_con[mug]         0.0   0.0   0.0   1.0  0.0
        limit_con[pc]          5.0   5.0   4.0   1.0  0.0
        limit_con[pen]         1.0   0.0   0.0   1.0  0.0
        weight_con            20.0  19.0  20.0  19.0  0.0

        Notes
        -----

        - If the :meth:`Model.solve` method is used with :code:`frame=True`
          parameter, the MILP solver returns multiple solutions.
          You can retreive different results by using the :code:`solution`
          parameter.

        """
        if vtype == 'Primal' or vtype == 'primal':
            if pivot:
                return self._primalSolution.pivot_table(
                    index=['var'], columns=['solution'], values='value')
            elif solution and 'solution' in self._primalSolution:
                return self._primalSolution.loc[
                    self._primalSolution['solution'] == solution]
            else:
                return self._primalSolution
        elif vtype == 'Dual' or vtype == 'dual':
            if pivot:
                return self._dualSolution.pivot_table(
                    index=['con'], columns=['solution'], values='value')
            elif solution and 'solution' in self._dualSolution:
                return self._dualSolution.loc[
                    self._dualSolution['solution'] == solution]
            else:
                return self._dualSolution
        else:
            raise ValueError('Solution type should be \'primal\' or \'dual\'')

    def get_tuner_results(self):
        """
        Returns the tuning results

        Examples
        --------

        >>> m.tune_parameters(tunerParameters={'maxConfigs': 10})
        >>> results = m.get_tuner_reults()

        Returns
        -------
        tunerResults : dict
           Returns tuner results as a dictionary.

           Its members are

           - Performance Information
           - Tuner Information
           - Tuner Summary
           - Tuner Results

        See also
        --------
        :meth:`Model.tune_parameters`

        """
        return self._tunerResults

    def set_session(self, session):
        """
        Sets the session of model

        Parameters
        ----------
        session : :class:`swat.cas.connection.CAS` or \
:class:`saspy.SASsession`
            CAS or SAS Session object

        Notes
        -----

        * You can use CAS sessions (via SWAT package) or SAS sessions (via SASPy package)
        * Session of a model can be set at initialization.
          See :class:`Model`.

        """
        self._session = session

    def get_session(self):
        """
        Returns the session of the model

        Returns
        -------
        session : :class:`swat.cas.connection.CAS` or \
:class:`saspy.SASsession`
            Session of the model, or None
        """
        return self._session

    def get_sets(self):
        """
        Returns a list of :class:`Set` objects in the model

        Returns
        -------
        set_list : list
            List of sets in the model

        Examples
        --------

        >>> m.get_sets()
        [sasoptpy.abstract.Set(name=W, settype=['str', 'num']), sasoptpy.abstract.Set(name=I, settype=['num']), sasoptpy.abstract.Set(name=J, settype=['num'])]

        """
        return list(self._setDict.values())

    def get_parameters(self):
        """
        Returns a list of :class:`abstract.Parameter` and
        :class:`abstract.ParameterGroup` objects in the model

        Returns
        -------
        param_list : list
            List of parameters in the model

        Examples
        --------

        >>> for i in m.get_parameters():
        ...     print(i.get_name(), type(i))
        p <class 'sasoptpy.abstract.parameter_group.ParameterGroup'>
        r <class 'sasoptpy.abstract.parameter.Parameter'>

        """
        return list(self._parameterDict.values())

    def get_statements(self):
        """
        Returns a list of all statements inside the model

        Returns
        -------
        st_list : list
            List of all statement objects

        Examples
        --------

        >>> m.add_statement(so.abstract.LiteralStatement("expand;"))
        >>> m.get_statements()
        [<sasoptpy.abstract.statement.literal.LiteralStatement object at 0x7fe0202fc358>]
        >>> print(m.to_optmodel())
        proc optmodel;
        var x;
        min obj1 = x * x;
        expand;
        solve;
        quit;

        """
        return list(self._statementDict.values())

    def get_implicit_variables(self):
        """
        Returns a list of implicit variables

        Returns
        -------
        implicit_variables : list
            List of implicit variables in the model

        Examples
        --------

        >>> m = so.Model(name='test_add_impvar')
        >>> x = m.add_variables(5, name='x')
        >>> y = m.add_implicit_variable((i * x[i] + x[i] ** 2 for i in range(5)),
                                    name='y')
        >>> assertEqual([y], m.get_implicit_variables())
        True

        """
        return list(self._impvarDict.values())

    def _get_dropped_cons(self):
        return self._droppedCons

    def _get_dropped_vars(self):
        return self._droppedVars

    def print_solution(self):
        """
        Prints the current values of the variables

        Examples
        --------

        >>> m.solve()
        >>> m.print_solution()
        x: 2.0
        y: 0.0

        See also
        --------
        :func:`Model.get_solution`

        Notes
        -----

        - This function might not work for abstract variables and nonlinear
          models.

        """
        for v in self.loop_variables():
            print('{}: {}'.format(v.get_name(), v._value))

    def to_frame(self, **kwargs):
        warnings.warn('Use to_mps for obtaining problem in MPS format',
                      DeprecationWarning)
        self.to_mps(**kwargs)

    def to_mps(self, **kwargs):
        """
        Returns the problem in MPS format

        Examples
        --------

        >>> print(n.to_mps())
            Field1 Field2 Field3  Field4 Field5  Field6  _id_
        0     NAME             n     0.0            0.0     1
        1     ROWS                   NaN            NaN     2
        2      MIN  myobj            NaN            NaN     3
        3  COLUMNS                   NaN            NaN     4
        4               y  myobj     2.0            NaN     5
        5      RHS                   NaN            NaN     6
        6   RANGES                   NaN            NaN     7
        7   BOUNDS                   NaN            NaN     8
        8       FR    BND      y     NaN            NaN     9
        9   ENDATA                   0.0            0.0    10

        """
        return sasoptpy.interface.to_mps(self, **kwargs)

    def export_mps(self, filename=None, fetch=False, **kwargs):
        """
        Exports model in MPS format

        Examples
        --------

        - Writing a linear optimization or mixed integer linear optimization model into MPS file
          
          >>> m.export_mps('my_problem.mps')

        - Returning the MPS string as a Python variable
          
          >>> m.export_mps(fetch=True)

        """
        if self._is_linear():
            mps = sasoptpy.util.export_to_mps(self, filename=filename, **kwargs)
            if fetch:
                return mps
        else:
            raise ValueError("Model is linear or has abstract components")

    def to_optmodel(self, **kwargs):
        """
        Returns the model in OPTMODEL format

        Examples
        --------

        >>> print(n.to_optmodel())
        proc optmodel;
        var y init 2;
        min myobj = 2 * y;
        solve;
        quit;

        """
        return sasoptpy.interface.to_optmodel(self, **kwargs)

    def __str__(self):
        """
        Returns a string representation of the Model object.

        Examples
        --------

        >>> print(m)
        Model: [
          Name: knapsack
          Session: cashost:casport
          Objective: MAX [8 * get[clock] + 10 * get[mug] + 15 * get[headphone]\
 + 20 * get[book] + get[pen]]
          Variables (5): [
            get[clock]
            get[mug]
            get[headphone]
            get[book]
            get[pen]
          ]
          Constraints (6): [
            get[clock] <=  3
            get[mug] <=  5
            get[headphone] <=  2
            get[book] <=  10
            get[pen] <=  15
            4 * get[clock] + 6 * get[mug] + 7 * get[headphone] + \
12 * get[book] + get[pen] <=  55
          ]
        ]

        """
        s = 'Model: [\n'
        s += '  Name: {}\n'.format(self.get_name())
        if self._session is not None:
            s += '  Session: {}:{}\n'.format(self._session._hostname,
                                             self._session._port)
        s += '  Objective: {} [{}]\n'.format(self.get_objective().get_sense(),
                                             self.get_objective())
        s += '  Variables ({}): [\n'.format(len(self.get_variables()))
        for i in self.loop_variables():
            s += '    {}\n'.format(i)
        s += '  ]\n'
        s += '  Constraints ({}): [\n'.format(len(self.get_constraints()))
        for i in self.loop_constraints():
            s += '    {}\n'.format(i)
        s += '  ]\n'
        s += ']'
        return s

    def __repr__(self):
        """
        Returns a representation of the Model object.

        Examples
        --------

        >>> print(repr(m))
        sasoptpy.Model(name='model1', session=CAS('cashost', 12345,
            'username', protocol='cas', name='py-session-1',
            session='594ad8d5-6a7b-3443-a155-be59177e8d23'))

        """
        if self._session is not None:
            stype = self.get_session_type()
            if stype == 'SAS':
                s = "sasoptpy.Model(name='{}', session=saspy.SASsession(cfgname='{}'))".format(
                         self.get_name(), self._session.sascfg.name)
            elif stype == 'CAS':
                s = 'sasoptpy.Model(name=\'{}\', session={})'.format(
                    self.get_name(), self._session)
            else:
                raise TypeError('Invalid session type: {}'.format(type(self.get_session())))
        else:
            s = 'sasoptpy.Model(name=\'{}\')'.format(self.get_name())
        return s

    def _defn(self):
        s = 'problem {}'.format(self.get_name())
        vars = [s.get_name() for s in self.get_grouped_variables().values()]
        cons = [s.get_name() for s in self.get_grouped_constraints().values()]
        obj = self.get_objective()
        objs = []
        if not obj.is_default():
            objs.append(obj.get_name())
        elements = ' '.join(vars + cons + objs)
        if elements != '':
            s += ' include ' + elements
        s += ';'
        return s

    def _expr(self):
        return self.to_optmodel()

    def _is_linear(self):
        """
        Checks whether the model can be written as a linear model (in MPS format)

        Returns
        -------
        is_linear : boolean
            True if model does not have any nonlinear components or abstract\
            operations, False otherwise
        """
        for c in self.loop_constraints():
            if not c._is_linear():
                return False
        if not self._objective._is_linear():
            return False
        return True

    def _has_integer_vars(self):
        for v in self._variableDict.values():
            if v._type != sasoptpy.CONT:
                return True
        return False

    def get_session_type(self):
        """
        Tests whether the model session is defined and still active

        Returns
        -------
        session : string
            'CAS' for CAS sessions, 'SAS' for SAS sessions

        """
        # Check if session is defined
        return sasoptpy.util.get_session_type(self._session)

    @sasoptpy.containable
    def solve(self, **kwargs):
        """
        Solves the model by calling CAS or SAS optimization solvers

        Parameters
        ----------
        options : dict, optional
            Solver options as a dictionary object
        submit : boolean, optional
            When set to `True`, calls the solver
        name : string, optional
            Name of the table
        frame : boolean, optional
            When set to `True`, uploads the problem as a DataFrame in MPS format
        drop : boolean, optional
            When set to `True`, drops the MPS table after solve (only CAS)
        replace : boolean, optional
            When set to `True`, replaces an existing MPS table (only CAS and MPS)
        primalin : boolean, optional
            When set to `True`, uses initial values (only MILP)
        verbose : boolean, optional (experimental)
            When set to `True`, prints the generated OPTMODEL code

        Returns
        -------
        solution : :class:`pandas.DataFrame`
            Solution of the optimization model

        Examples
        --------

        >>> m.solve()
        NOTE: Initialized model food_manufacture_1
        NOTE: Converting model food_manufacture_1 to DataFrame
        NOTE: Added action set 'optimization'.
        ...
        NOTE: Optimal.
        NOTE: Objective = 107842.59259.
        NOTE: The Dual Simplex solve time is 0.01 seconds.

        >>> m.solve(options={'maxtime': 600})

        >>> m.solve(options={'algorithm': 'ipm'})

        Notes
        -----

        * Some of the options listed under the ``options`` argument might not be
          passed, depending on which CAS action is being used.
        * The ``option`` argument should be a dictionary, where keys are
          option names. For example, ``m.solve(options={'maxtime': 600})``
          limits the solution time to 600 seconds.
        * See :ref:`solver-options` for a list of solver options.


        """

        return sasoptpy.util.submit_for_solve(self, **kwargs)

    def tune_parameters(self, **kwargs):
        """
        Tunes the model to find ideal solver parameters

        Parameters
        ----------
        kwargs :
           Keyword arguments as defined in the optimization.tuner action.

           Acceptable values are:

           - `milpParameters <https://go.documentation.sas.com/?docsetId=casactmopt&docsetTarget=cas-optimization-tuner.htm&docsetVersion=8.5&locale=en#PYTHON.cas-optimization-tuner-milpparameters>`_:
             Parameters for the solveMilp action, such as
             `maxTime`, `heuristics`, `feasTol`
           - `tunerParameters <https://go.documentation.sas.com/?docsetId=casactmopt&docsetTarget=cas-optimization-tuner.htm&docsetVersion=8.5&locale=en#PYTHON.cas-optimization-tuner-tunerparameters>`_:
             Parameters for the tuner itself, such as
             `maxConfigs`, `printLevel`, `logFreq`
           - `tuningParameters <https://go.documentation.sas.com/?docsetId=casactmopt&docsetTarget=cas-optimization-tuner.htm&docsetVersion=8.5&locale=en#PYTHON.cas-optimization-tuner-tuningparameters>`_:
             List of parameters to be tuned, such as
             `cutStrategy`, `presolver`, `restarts`

        Returns
        -------
        tunerResults : :class:`swat.dataframe.SASDataFrame`
           Tuning results as a table

        Examples
        --------

        >>> m = so.Model(name='model1')
        >>> ...
        >>> results = m.tune_parameters(tunerParameters={'maxConfigs': 10})
        NOTE: Initialized model knapsack_with_tuner.
        NOTE: Added action set 'optimization'.
        NOTE: Uploading the problem DataFrame to the server.
        NOTE: Cloud Analytic Services made the uploaded file available as table KNAPSACK_WITH_TUNER in caslib CASUSER(casuser).
        NOTE: The table KNAPSACK_WITH_TUNER has been created in caslib CASUSER(casuser) from binary data uploaded to Cloud Analytic Services.
        NOTE: Start to tune the MILP
                 SolveCalls  Configurations    BestTime        Time
                          1               1        0.21        0.26
                          2               2        0.19        0.50
                          3               3        0.19        0.72
                          4               4        0.19        0.95
                          5               5        0.19        1.17
                          6               6        0.19        1.56
                          7               7        0.18        1.76
                          8               8        0.17        1.96
                          9               9        0.17        2.16
                         10              10        0.17        2.35
        NOTE: Configuration limit reached.
        NOTE: The tuning time is 2.35 seconds.
        >>> print(results)
           Configuration conflictSearch  ... Sum of Run Times Percentage Successful
        0            0.0      automatic  ...             0.20                 100.0
        1            1.0           none  ...             0.17                 100.0
        2            2.0           none  ...             0.17                 100.0
        3            3.0       moderate  ...             0.17                 100.0
        4            4.0           none  ...             0.18                 100.0
        5            5.0           none  ...             0.18                 100.0
        6            6.0     aggressive  ...             0.18                 100.0
        7            7.0       moderate  ...             0.18                 100.0
        8            8.0     aggressive  ...             0.19                 100.0
        9            9.0      automatic  ...             0.36                 100.0


        >>> results = m.tune_parameters(
               milpParameters={'maxtime': 10},
               tunerParameters={'maxConfigs': 20, 'logfreq': 5},
               tuningParameters=[
                  {'option': 'presolver', 'initial': 'none', 'values': ['basic', 'aggressive', 'none']},
                  {'option': 'cutStrategy'},
                  {'option': 'strongIter', 'initial': -1, 'values': [-1, 100, 1000]}
               ])
        NOTE: Added action set 'optimization'.
        NOTE: Uploading the problem DataFrame to the server.
        NOTE: Cloud Analytic Services made the uploaded file available as table KNAPSACK_WITH_TUNER in caslib CASUSER(casuser).
        NOTE: The table KNAPSACK_WITH_TUNER has been created in caslib CASUSER(casuser) from binary data uploaded to Cloud Analytic Services.
        NOTE: Start to tune the MILP
                 SolveCalls  Configurations    BestTime        Time
                          5               5        0.17        1.01
                         10              10        0.17        2.00
                         15              15        0.17        2.98
                         20              20        0.17        3.95
        NOTE: Configuration limit reached.
        NOTE: The tuning time is 3.95 seconds.
        >>> print(results)
            Configuration conflictSearch  ... Sum of Run Times Percentage Successful
        0             0.0      automatic  ...             0.17                 100.0
        1             1.0           none  ...             0.16                 100.0
        2             2.0           none  ...             0.16                 100.0
        3             3.0           none  ...             0.16                 100.0
        4             4.0           none  ...             0.16                 100.0
        5             5.0           none  ...             0.17                 100.0
        6             6.0           none  ...             0.17                 100.0
        7             7.0           none  ...             0.17                 100.0
        8             8.0           none  ...             0.17                 100.0
        9             9.0           none  ...             0.17                 100.0
        10           10.0           none  ...             0.17                 100.0
        11           11.0     aggressive  ...             0.17                 100.0
        12           12.0           none  ...             0.17                 100.0
        13           13.0     aggressive  ...             0.17                 100.0
        14           14.0      automatic  ...             0.17                 100.0
        15           15.0           none  ...             0.17                 100.0
        16           16.0           none  ...             0.17                 100.0
        17           17.0       moderate  ...             0.17                 100.0
        18           18.0       moderate  ...             0.17                 100.0
        19           19.0           none  ...             0.17                 100.0

        Notes
        -----
        - See `SAS Optimization documentation
          <https://go.documentation.sas.com/?docsetId=casactmopt&docsetTarget=cas-optimization-tuner.htm&docsetVersion=8.5&locale=en#PYTHON.cas-optimization-tuner-tunerparameters>`_
          for a full list of tunable parameters.
        - See `Optimization Action Set documentation
          <https://go.documentation.sas.com/?docsetId=casactmopt&docsetTarget=casactmopt_optimization_details35.htm&docsetVersion=8.5&locale=en>`_.

        See also
        --------
        :meth:`Model.get_tuner_results`
        """
        return sasoptpy.util.submit_for_tune(self, **kwargs)

    def _set_abstract_values(self, name, value):
        """
        Searches for the missing/abstract variable names and set their values
        """
        original_name = sasoptpy.util.get_group_name(name)
        group = self.get_variable(original_name)
        if group is not None:
            v = group.get_member_by_name(name)
            v.set_value(value)

    def clear_solution(self):
        """
        Clears the cached solution of the model

        Notes
        -----
        - This method cleans the optimal objective value and solution time
          parameters of the model.
        """
        self._objval = None
        self.response = None
        self._soltime = 0

